#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import annotations

import binascii
import enum
import errno
import json
import sys
import typing

from .opener import Opener, FileOpener

_VALID_VERSIONS = ["v0"]

# alias open to _open to support test assertions when running
# as UID 0
_open = open

PasswdEntryTuple = typing.Tuple[str, str, str, str, str, str, str]
GroupEntryTuple = typing.Tuple[str, str, str, str]

# JSONData is not really a completely valid representation of json in
# the type system, but it's good enough for now. Using it can help
# clarify what works with json and what works with real python dicts.
JSONData = dict[str, typing.Any]

# the standard location for samba's smb.conf
SMB_CONF = "/etc/samba/smb.conf"
CLUSTER_META_JSON = "/var/lib/ctdb/shared/ctdb-nodes.json"
CTDB_NODES_PATH = "/var/lib/ctdb/shared/nodes"
CTDB_RECLOCK = "/var/lib/ctdb/shared/RECOVERY"

CTDB: typing.Final[str] = "ctdb"
ADDC: typing.Final[str] = "addc"
FEATURES: typing.Final[str] = "instance_features"

# cache for json schema data
_JSON_SCHEMA: dict[str, typing.Any] = {}


class ConfigFormat(enum.Enum):
    JSON = "json"
    TOML = "toml"
    YAML = "yaml"


class ValidationUnsupported(Exception):
    pass


class _FakeRefResolutionError(Exception):
    pass


class ConfigFormatUnsupported(Exception):
    pass


if sys.version_info >= (3, 11):

    def _load_toml(source: typing.IO) -> JSONData:
        try:
            import tomllib
        except ImportError:
            raise ConfigFormatUnsupported(ConfigFormat.TOML)
        return tomllib.load(source)

else:

    def _load_toml(source: typing.IO) -> JSONData:
        try:
            import tomli
        except ImportError:
            raise ConfigFormatUnsupported(ConfigFormat.TOML)
        if typing.TYPE_CHECKING:
            assert isinstance(source, typing.BinaryIO)
        return tomli.load(source)


def _load_yaml(source: typing.IO) -> JSONData:
    try:
        import yaml
    except ImportError:
        raise ConfigFormatUnsupported(ConfigFormat.YAML)
    return yaml.safe_load(source) or {}


def _detect_format(fname: str) -> ConfigFormat:
    if fname.endswith(".toml"):
        return ConfigFormat.TOML
    if fname.endswith((".yaml", ".yml")):
        return ConfigFormat.YAML
    return ConfigFormat.JSON


def _schema_validate(data: dict[str, typing.Any], version: str) -> None:
    try:
        import jsonschema  # type: ignore[import]
    except ImportError:
        raise ValidationUnsupported()
    try:
        _refreserror = getattr(jsonschema, "RefResolutionError")
    except AttributeError:
        _refreserror = _FakeRefResolutionError

    global _JSON_SCHEMA
    if version == "v0" and version not in _JSON_SCHEMA:
        try:
            import sambacc.schema.conf_v0_schema

            _JSON_SCHEMA[version] = sambacc.schema.conf_v0_schema.SCHEMA
        except ImportError:
            raise ValidationUnsupported()
    try:
        jsonschema.validate(instance=data, schema=_JSON_SCHEMA[version])
    except _refreserror:
        raise ValidationUnsupported()


def _check_config_version(data: JSONData) -> str:
    """Return the config data or raise a ValueError if the config
    is invalid or incomplete.
    """
    # short-cut to validate that this is something we want to consume
    version = data.get("samba-container-config")
    if version is None:
        raise ValueError("Invalid config: no samba-container-config key")
    elif version not in _VALID_VERSIONS:
        raise ValueError(f"Invalid config: unknown version {version}")
    return version


def _check_config_valid(
    data: JSONData, version: str, required: typing.Optional[bool] = None
) -> None:
    if required or required is None:
        try:
            _schema_validate(data, version)
        except ValidationUnsupported:
            if required:
                raise


def read_config_files(
    fnames: list[str],
    *,
    require_validation: typing.Optional[bool] = None,
    opener: typing.Optional[Opener] = None,
) -> GlobalConfig:
    """Read the global container config from the given filenames.
    At least one of the files from the fnames list must exist and contain
    a valid config. If none of the file names exist an error will be raised.
    """
    # NOTE: Right now if more than one config exists they'll be "merged" but
    # the merging is very simplistic right now. Mainly we expect that the
    # users will be split from the main config (for security reasons) but
    # it would be nicer to have a good merge algorithm handle everything
    # smarter at some point.
    opener = opener or FileOpener()
    gconfig = GlobalConfig()
    readfiles = set()
    for fname in fnames:
        config_format = _detect_format(str(fname))
        try:
            with opener.open(fname) as fh:
                gconfig.load(
                    fh,
                    require_validation=require_validation,
                    config_format=config_format,
                )
            readfiles.add(fname)
        except OSError as err:
            if getattr(err, "errno", 0) != errno.ENOENT:
                raise
    if not readfiles:
        # we read nothing! don't proceed
        raise ValueError(f"None of the config file paths exist: {fnames}")
    return gconfig


class SambaConfig(typing.Protocol):
    """Minimal samba configuration protocol."""

    def global_options(self) -> typing.Iterable[typing.Tuple[str, str]]:
        """Return global options for Samba."""
        ...  # pragma: no cover

    def shares(self) -> typing.Iterable[ShareConfig]:
        """Return share configurations for Samba."""
        ...  # pragma: no cover


class GlobalConfig:
    def __init__(
        self,
        source: typing.Optional[typing.IO] = None,
        *,
        initial_data: typing.Optional[JSONData] = None,
    ) -> None:
        self.data: JSONData = {} if initial_data is None else initial_data
        if source is not None:
            self.load(source)

    def load(
        self,
        source: typing.IO,
        *,
        require_validation: typing.Optional[bool] = None,
        config_format: typing.Optional[ConfigFormat] = None,
    ) -> None:
        config_format = config_format or ConfigFormat.JSON
        if config_format == ConfigFormat.TOML:
            data = _load_toml(source)
        elif config_format == ConfigFormat.YAML:
            data = _load_yaml(source)
        else:
            data = json.load(source)
        _check_config_valid(
            data, _check_config_version(data), require_validation
        )
        self.data.update(data)

    def get(self, ident: str) -> InstanceConfig:
        iconfig = self.data["configs"][ident]
        return InstanceConfig(self, iconfig)


class InstanceConfig:
    def __init__(self, conf: GlobalConfig, iconfig: dict):
        self.gconfig = conf
        self.iconfig = iconfig

    def global_options(self) -> typing.Iterable[typing.Tuple[str, str]]:
        """Iterate over global options."""
        # Pull in all global sections that apply
        try:
            gnames = self.iconfig["globals"]
        except KeyError:
            # no globals section in the instance means no global options
            return
        for gname in gnames:
            global_section = self.gconfig.data["globals"][gname]
            for k, v in global_section.get("options", {}).items():
                yield k, v
        # Special, per-instance settings
        instance_name = self.iconfig.get("instance_name", None)
        if instance_name:
            yield "netbios name", instance_name

    def uid_base(self) -> int:
        return 1000

    def gid_base(self) -> int:
        return 1000

    def shares(self) -> typing.Iterable[ShareConfig]:
        """Iterate over share configs."""
        for sname in self.iconfig.get("shares", []):
            yield ShareConfig(self.gconfig, sname, iconfig=self.iconfig)

    def users(self) -> typing.Iterable[UserEntry]:
        all_users = self.gconfig.data.get("users", {}).get("all_entries", {})
        for n, entry in enumerate(all_users):
            yield UserEntry(self, entry, n)

    def groups(self) -> typing.Iterable[GroupEntry]:
        user_gids = {u.gid: u for u in self.users()}
        all_groups = self.gconfig.data.get("groups", {}).get("all_entries", {})
        for n, entry in enumerate(all_groups):
            ge = GroupEntry(self, entry, n)
            if ge.gid in user_gids:
                del user_gids[ge.gid]
            yield ge
        for uentry in user_gids.values():
            yield uentry.vgroup()

    @property
    def with_ctdb(self) -> bool:
        return CTDB in self.iconfig.get(FEATURES, [])

    @property
    def with_addc(self) -> bool:
        return ADDC in self.iconfig.get(FEATURES, [])

    def ctdb_smb_config(self) -> CTDBSambaConfig:
        if not self.with_ctdb:
            raise ValueError("ctdb not supported in configuration")
        return CTDBSambaConfig()

    def ctdb_config(self) -> dict[str, str]:
        """Common configuration of CTDB itself."""
        if not self.with_ctdb:
            return {}
        ctdb = dict(self.gconfig.data.get("ctdb", {}))
        ctdb.setdefault("cluster_meta_uri", CLUSTER_META_JSON)
        ctdb.setdefault("nodes_path", CTDB_NODES_PATH)
        ctdb.setdefault("recovery_lock", CTDB_RECLOCK)
        ctdb.setdefault("log_level", "NOTICE")
        ctdb.setdefault("script_log_level", "ERROR")
        ctdb.setdefault("realtime_scheduling", "false")
        # this whole thing really needs to be turned into a real object type
        ctdb.setdefault("public_addresses", [])
        return ctdb

    def domain(self) -> DomainConfig:
        """Return the general domain settings for this DC instance."""
        if not self.with_addc:
            raise ValueError("ad dc not supported by configuration")
        domains = self.gconfig.data.get("domain_settings", {})
        instance_name: str = self.iconfig.get("instance_name", "")
        return DomainConfig(
            drec=domains[self.iconfig["domain_settings"]],
            instance_name=instance_name,
        )

    def domain_users(self) -> typing.Iterable[DomainUserEntry]:
        if not self.with_addc:
            raise ValueError("ad dc not supported by configuration")
        ds_name: str = self.iconfig["domain_settings"]
        dusers = self.gconfig.data.get("domain_users", {}).get(ds_name, [])
        for n, entry in enumerate(dusers):
            yield DomainUserEntry(self, entry, n)

    def domain_groups(self) -> typing.Iterable[DomainGroupEntry]:
        if not self.with_addc:
            raise ValueError("ad dc not supported by configuration")
        ds_name: str = self.iconfig["domain_settings"]
        dgroups = self.gconfig.data.get("domain_groups", {}).get(ds_name, [])
        for n, entry in enumerate(dgroups):
            yield DomainGroupEntry(self, entry, n)

    def organizational_units(self) -> typing.Iterable[OrganizationalUnitEntry]:
        if not self.with_addc:
            raise ValueError("ad dc not supported by configuration")
        ds_name: str = self.iconfig["domain_settings"]
        o_units = self.gconfig.data.get("organizational_units", {}).get(
            ds_name, []
        )
        for n, entry in enumerate(o_units):
            yield OrganizationalUnitEntry(self, entry, n)

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, InstanceConfig) and self.iconfig == other.iconfig:
            self_shares = _shares_data(self.gconfig, self.iconfig)
            other_shares = _shares_data(other.gconfig, other.iconfig)
            self_globals = _globals_data(self.gconfig, self.iconfig)
            other_globals = _globals_data(other.gconfig, other.iconfig)
            return (
                self_shares == other_shares and self_globals == other_globals
            )
        return False


class CTDBSambaConfig:
    def global_options(self) -> typing.Iterable[typing.Tuple[str, str]]:
        return [
            ("clustering", "yes"),
            ("ctdb:registry.tdb", "yes"),
            ("include", "registry"),
        ]

    def shares(self) -> typing.Iterable[ShareConfig]:
        return []


class ShareConfig:
    def __init__(
        self,
        conf: GlobalConfig,
        sharename: str,
        iconfig: typing.Optional[dict] = None,
    ) -> None:
        self.gconfig = conf
        self.name = sharename
        self.iconfig = iconfig or {}

    def share_options(self) -> typing.Iterable[typing.Tuple[str, str]]:
        """Iterate over share options."""
        share_section = self.gconfig.data["shares"][self.name]
        return iter(share_section.get("options", {}).items())

    def path(self) -> typing.Optional[str]:
        """Return the path value from the smb.conf options."""
        share_section = self.gconfig.data["shares"][self.name]
        try:
            return share_section["options"]["path"]
        except KeyError:
            return None

    def permissions_config(self) -> PermissionsConfig:
        """Return a permissions configuration for the share."""
        # each share can have it's own permissions config,
        # but if it does not it will default to the instance's
        # config
        try:
            share_perms = self.gconfig.data["shares"][self.name]["permissions"]
            return PermissionsConfig(share_perms)
        except KeyError:
            pass
        try:
            instance_perms = self.iconfig["permissions"]
            return PermissionsConfig(instance_perms)
        except KeyError:
            pass
        # use the internal defaults
        return PermissionsConfig({})


class UserEntry:
    def __init__(self, iconf: InstanceConfig, urec: dict, num: int):
        self.iconfig = iconf
        self.username = urec["name"]
        self.entry_num = num
        self._uid = urec.get("uid")
        self._gid = urec.get("gid")
        self._nt_passwd = str(urec.get("nt_hash", ""))
        self._plaintext_passwd = str(urec.get("password", ""))
        if self._uid is not None:
            if not isinstance(self._uid, int):
                raise ValueError("invalid uid value")
        if self._gid is not None:
            if not isinstance(self._gid, int):
                raise ValueError("invalid gid value")

    @property
    def uid(self) -> int:
        if self._uid:
            return self._uid
        return self.iconfig.uid_base() + self.entry_num

    @property
    def gid(self) -> int:
        if self._gid:
            return self._gid
        return self.iconfig.gid_base() + self.entry_num

    @property
    def dir(self) -> str:
        return "/invalid"

    @property
    def shell(self) -> str:
        return "/bin/false"

    @property
    def nt_passwd(self) -> bytes:
        # the json will store the hash as a hex encoded string
        return binascii.unhexlify(self._nt_passwd)

    @property
    def plaintext_passwd(self) -> str:
        return self._plaintext_passwd

    def passwd_fields(self) -> PasswdEntryTuple:
        # fields: name, passwd, uid, gid, GECOS, dir, shell
        return (
            self.username,
            "x",
            str(self.uid),
            str(self.gid),
            "",
            self.dir,
            self.shell,
        )

    def vgroup(self) -> GroupEntry:
        """In case there is no explicit group for the specified user. This
        handy method makes a "virtual" group based on the user info.
        """
        return GroupEntry(
            self.iconfig, dict(name=self.username, gid=self.gid), 0
        )


class GroupEntry:
    def __init__(self, iconf: InstanceConfig, grec: dict, num: int):
        self.iconfig = iconf
        self.groupname = grec["name"]
        self.entry_num = num
        self._gid = grec.get("gid")
        self.ou = grec.get("ou")
        if self._gid is not None:
            if not isinstance(self._gid, int):
                raise ValueError("invalid gid value")

    @property
    def gid(self) -> int:
        if self._gid:
            return self._gid
        return self.iconfig.gid_base() + self.entry_num

    def group_fields(self) -> GroupEntryTuple:
        # fields: name, passwd, gid, members(comma separated)
        return (self.groupname, "x", str(self.gid), "")


class DomainConfig:
    def __init__(self, drec: dict, instance_name: str):
        self.realm = drec["realm"]
        self.short_domain = drec.get("short_domain", "")
        self.admin_password = drec.get("admin_password", "")
        self.interface_config = DCInterfaceConfig(drec.get("interfaces", {}))
        self.dcname = instance_name


class DCInterfaceConfig:
    def __init__(self, iface_config: dict) -> None:
        self.include_pattern: str = iface_config.get("include_pattern", "")
        self.exclude_pattern: str = iface_config.get("exclude_pattern", "")

    @property
    def configured(self) -> bool:
        """Return true if at least one interface property has been set."""
        return bool(self.include_pattern) or bool(self.exclude_pattern)


class DomainUserEntry(UserEntry):
    def __init__(self, iconf: InstanceConfig, urec: dict, num: int):
        super().__init__(iconf, urec, num)
        self.surname = urec.get("surname")
        self.given_name = urec.get("given_name")
        self.member_of = urec.get("member_of", [])
        self.ou = urec.get("ou")
        if not isinstance(self.member_of, list):
            raise ValueError("member_of should contain a list of group names")


class DomainGroupEntry(GroupEntry):
    pass


class OrganizationalUnitEntry:
    def __init__(self, iconf: InstanceConfig, urec: dict, num: int):
        self.iconfig = iconf
        self.ou_name = urec["name"]
        self.entry_num = num


class PermissionsConfig:
    _method_key: str = "method"
    _status_xattr_key: str = "status_xattr"
    _default_method: str = "none"
    _default_status_xattr: str = "user.share-perms-status"

    def __init__(self, pconf: dict[str, str]) -> None:
        self._pconf = pconf
        self.method: str = pconf.get(self._method_key, self._default_method)
        self.status_xattr: str = pconf.get(
            self._status_xattr_key, self._default_status_xattr
        )

    @property
    def options(self) -> dict[str, str]:
        filter_keys = {self._method_key, self._status_xattr_key}
        return {k: v for k, v in self._pconf.items() if k not in filter_keys}


def _shares_data(gconfig: GlobalConfig, iconfig: dict) -> list:
    try:
        shares = iconfig["shares"]
    except KeyError:
        return []
    return [gconfig.data["shares"][n] for n in shares]


def _globals_data(gconfig: GlobalConfig, iconfig: dict) -> list:
    try:
        gnames = iconfig["globals"]
    except KeyError:
        return []
    return [gconfig.data["globals"][n] for n in gnames]



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing

from .textfile import TextFileLoader
from sambacc import config


class LineFileLoader(TextFileLoader):
    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.lines: list[str] = []

    def loadlines(self, lines: typing.Iterable[str]) -> None:
        """Load in the lines from the text source."""
        for line in lines:
            self.lines.append(line)

    def dumplines(self) -> typing.Iterable[str]:
        """Dump the file content as lines of text."""
        prev = None
        for line in self.lines:
            if prev and not prev.endswith("\n"):
                yield "\n"
            yield line
            prev = line


class PasswdFileLoader(LineFileLoader):
    def __init__(self, path: str = "/etc/passwd") -> None:
        super().__init__(path)
        self._usernames: set[str] = set()

    def readfp(self, fp: typing.IO) -> None:
        super().readfp(fp)
        self._update_usernames_cache()

    def _update_usernames_cache(self) -> None:
        for line in self.lines:
            if ":" in line:
                u = line.split(":")[0]
                self._usernames.add(u)

    def add_user(self, user_entry: config.UserEntry) -> None:
        if user_entry.username in self._usernames:
            return
        line = "{}\n".format(":".join(user_entry.passwd_fields()))
        self.lines.append(line)
        self._usernames.add(user_entry.username)


class GroupFileLoader(LineFileLoader):
    def __init__(self, path: str = "/etc/group") -> None:
        super().__init__(path)
        self._groupnames: set[str] = set()

    def readfp(self, fp: typing.IO) -> None:
        super().readfp(fp)
        self._update_groupnames_cache()

    def _update_groupnames_cache(self) -> None:
        for line in self.lines:
            if ":" in line:
                u = line.split(":")[0]
                self._groupnames.add(u)

    def add_group(self, group_entry: config.GroupEntry) -> None:
        if group_entry.groupname in self._groupnames:
            return
        line = "{}\n".format(":".join(group_entry.group_fields()))
        self.lines.append(line)
        self._groupnames.add(group_entry.groupname)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import enum
import errno
import json
import logging
import subprocess
import typing

from .opener import Opener, FileOpener
from sambacc import samba_cmds
from sambacc.simple_waiter import Waiter

_logger = logging.getLogger(__name__)


class JoinError(Exception):
    def __init__(self, v: typing.Any) -> None:
        super().__init__(v)
        self.errors: list[typing.Any] = []


_PROMPT = object()
_PT = typing.TypeVar("_PT")
_PW = typing.Union[str, _PT]


class JoinBy(enum.Enum):
    PASSWORD = "password"
    FILE = "file"
    INTERACTIVE = "interactive"


class UserPass:
    """Encapsulate a username/password pair."""

    username: str = "Administrator"
    password: typing.Optional[_PW] = None

    def __init__(
        self,
        username: typing.Optional[str] = None,
        password: typing.Optional[_PW] = None,
    ) -> None:
        if username is not None:
            self.username = username
        if password is not None:
            self.password = password


class _JoinSource(typing.NamedTuple):
    method: JoinBy
    upass: typing.Optional[UserPass]
    path: str


class Joiner:
    """Utility class for joining to AD domain.

    Use the `add_source` method to add one or more sources of join auth
    data. Call `join` to commit and join the "host" to AD.
    """

    _net_ads_join = samba_cmds.net["ads", "join"]

    def __init__(
        self,
        marker: typing.Optional[str] = None,
        *,
        opener: typing.Optional[Opener] = None,
    ) -> None:
        self._source_paths: list[str] = []
        self._sources: list[_JoinSource] = []
        self.marker = marker
        self._opener = opener or FileOpener()

    def add_source(
        self,
        method: JoinBy,
        value: typing.Optional[typing.Union[str, UserPass]] = None,
    ) -> None:
        if method in {JoinBy.PASSWORD, JoinBy.INTERACTIVE}:
            if not isinstance(value, UserPass):
                raise ValueError("expected UserPass value")
            if method == JoinBy.PASSWORD:
                self.add_pw_source(value)
            else:
                self.add_interactive_source(value)
        elif method in {JoinBy.FILE}:
            if not isinstance(value, str):
                raise ValueError("expected str value")
            self.add_file_source(value)
        else:
            raise ValueError(f"invalid method: {method}")

    def add_file_source(self, path_or_uri: str) -> None:
        self._sources.append(_JoinSource(JoinBy.FILE, None, path_or_uri))

    def add_pw_source(self, value: UserPass) -> None:
        self._sources.append(_JoinSource(JoinBy.PASSWORD, value, ""))

    def add_interactive_source(self, value: UserPass) -> None:
        self._sources.append(_JoinSource(JoinBy.INTERACTIVE, value, ""))

    def join(self, dns_updates: bool = False) -> None:
        if not self._sources:
            raise JoinError("no sources for join data")
        errors = []
        for src in self._sources:
            try:
                if src.method is JoinBy.PASSWORD:
                    assert src.upass
                    upass = src.upass
                elif src.method is JoinBy.FILE:
                    assert src.path
                    upass = self._read_from(src.path)
                elif src.method is JoinBy.INTERACTIVE:
                    assert src.upass
                    upass = UserPass(src.upass.username, _PROMPT)
                else:
                    raise ValueError(f"invalid method: {src.method}")
                self._join(upass, dns_updates=dns_updates)
                self._set_marker()
                return
            except JoinError as join_err:
                errors.append(join_err)
        if errors:
            if len(errors) == 1:
                raise errors[0]
            err = JoinError("failed {} join attempts".format(len(errors)))
            err.errors = errors
            raise err

    def _read_from(self, path: str) -> UserPass:
        try:
            with self._opener.open(path) as fh:
                data = json.load(fh)
        except FileNotFoundError:
            raise JoinError(f"source file not found: {path}")
        except OSError as err:
            if getattr(err, "errno", 0) != errno.ENOENT:
                raise
            raise JoinError(f"resource not found: {path}")
        upass = UserPass()
        try:
            upass.username = data["username"]
            upass.password = data["password"]
        except KeyError as err:
            raise JoinError(f"invalid file content: {err}")
        if not isinstance(upass.username, str):
            raise JoinError("invalid file content: invalid username")
        if not isinstance(upass.password, str):
            raise JoinError("invalid file content: invalid password")
        return upass

    def _interactive_input(self) -> typing.Optional[typing.IO]:
        return None

    def _join(self, upass: UserPass, dns_updates: bool = False) -> None:
        args = []
        if not dns_updates:
            args.append("--no-dns-updates")
        args.extend(["-U", upass.username])

        if upass.password is _PROMPT:
            cmd = list(self._net_ads_join[args])
            proc = subprocess.Popen(cmd, stdin=self._interactive_input())
        else:
            cmd = list(self._net_ads_join[args])
            proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
            pw_data = samba_cmds.encode(upass.password)
            # mypy can't seem to handle the following lines, and none of my web
            # searches turned up a clear answer. ignore for now
            proc.stdin.write(pw_data)  # type: ignore
            proc.stdin.write(b"\n")  # type: ignore
            proc.stdin.close()  # type: ignore
        ret = proc.wait()
        if ret != 0:
            raise JoinError("failed to run {}".format(cmd))

    def _set_marker(self) -> None:
        if self.marker is not None:
            with open(self.marker, "w") as fh:
                json.dump({"joined": True}, fh)

    def did_join(self) -> bool:
        """Return true if the join marker exists and contains a true
        value in the joined key.
        """
        if self.marker is None:
            return False
        try:
            with open(self.marker) as fh:
                data = json.load(fh)
        except (ValueError, OSError):
            return False
        try:
            return data["joined"]
        except (TypeError, KeyError):
            return False


def join_when_possible(
    joiner: Joiner,
    waiter: Waiter,
    error_handler: typing.Optional[typing.Callable] = None,
) -> None:
    while True:
        if joiner.did_join():
            _logger.info("found valid join marker")
            return
        try:
            joiner.join()
            _logger.info("successful join")
            return
        except JoinError as err:
            if error_handler is not None:
                error_handler(err)
            else:
                raise
        waiter.wait()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing


class ConfigStore(typing.Protocol):
    def __getitem__(self, name: str) -> list[tuple[str, str]]:
        """Get an item, returning a config section."""
        ...  # pragma: no cover

    def __setitem__(self, name: str, value: list[tuple[str, str]]) -> None:
        """Set a new config section."""
        ...  # pragma: no cover

    def __iter__(self) -> typing.Iterator[str]:
        """Iterate over config sections in the store."""
        ...  # pragma: no cover


class SimpleConfigStore:
    def __init__(self) -> None:
        self._data: dict[str, list[tuple[str, str]]] = {}

    @property
    def writeable(self) -> bool:
        """True if using a read-write backend."""
        return True

    def __getitem__(self, name: str) -> list[tuple[str, str]]:
        return self._data[name]

    def __setitem__(self, name: str, value: list[tuple[str, str]]) -> None:
        self._data[name] = value

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._data.keys())

    def import_smbconf(
        self, src: ConfigStore, batch_size: typing.Optional[int] = None
    ) -> None:
        """Import content from one SMBConf configuration object into the
        current SMBConf configuration object.

        batch_size is ignored.
        """
        for sname in src:
            self[sname] = src[sname]


def write_store_as_smb_conf(out: typing.IO, conf: ConfigStore) -> None:
    """Write the configuration store in smb.conf format to `out`."""
    # unfortunately, AFAIK, there's no way for an smbconf to write
    # into a an smb.conf/ini style file. We have to do it on our own.
    # ---
    # Make sure global section comes first.
    sections = sorted(conf, key=lambda v: 0 if v == "global" else 1)
    for sname in sections:
        out.write(str("\n[{}]\n".format(sname)))
        for skey, sval in conf[sname]:
            out.write(str(f"\t{skey} = {sval}\n"))



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sys
import types
import importlib
import typing
import itertools

from sambacc.smbconf_api import ConfigStore


def _smbconf() -> types.ModuleType:
    return importlib.import_module("samba.smbconf")


def _s3smbconf() -> types.ModuleType:
    return importlib.import_module("samba.samba3.smbconf")


def _s3param() -> types.ModuleType:
    return importlib.import_module("samba.samba3.param")


if sys.version_info >= (3, 11):
    from typing import Self as _Self
else:
    _Self = typing.TypeVar("_Self", bound="SMBConf")


class SMBConf:
    """SMBConf wraps the samba smbconf library, supporting reading from and,
    when possible, writing to samba configuration backends.  The SMBConf type
    supports transactions using the context managager interface.  The SMBConf
    type can read and write configuration based on dictionary-like access,
    using shares as the keys. The global configuration is treated like a
    special "share".
    """

    def __init__(self, smbconf: typing.Any) -> None:
        self._smbconf = smbconf

    @classmethod
    def from_file(cls: typing.Type[_Self], path: str) -> _Self:
        """Open a smb.conf style configuration from the specified path."""
        return cls(_smbconf().init_txt(path))

    @classmethod
    def from_registry(
        cls: typing.Type[_Self],
        configfile: str = "/etc/samba/smb.conf",
        key: typing.Optional[str] = None,
    ) -> _Self:
        """Open samba's registry backend for configuration parameters."""
        s3_lp = _s3param().get_context()
        s3_lp.load(configfile)
        return cls(_s3smbconf().init_reg(key))

    @property
    def writeable(self) -> bool:
        """True if using a read-write backend."""
        return self._smbconf.is_writeable()

    # the extraneous `self: _Self` type makes mypy on python <3.11 happy.
    # otherwise it complains: `A function returning TypeVar should receive at
    # least one argument containing the same TypeVar`
    def __enter__(self: _Self) -> _Self:
        self._smbconf.transaction_start()
        return self

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, tb: typing.Any
    ) -> None:
        if exc_type is None:
            self._smbconf.transaction_commit()
            return
        return self._smbconf.transaction_cancel()

    def __getitem__(self, name: str) -> list[tuple[str, str]]:
        try:
            n2, values = self._smbconf.get_share(name)
        except _smbconf().SMBConfError as err:
            if err.error_code == _smbconf().SBC_ERR_NO_SUCH_SERVICE:
                raise KeyError(name)
            raise
        if name != n2:
            raise ValueError(f"section name invalid: {name!r} != {n2!r}")
        return values

    def __setitem__(self, name: str, value: list[tuple[str, str]]) -> None:
        try:
            self._smbconf.delete_share(name)
        except _smbconf().SMBConfError as err:
            if err.error_code != _smbconf().SBC_ERR_NO_SUCH_SERVICE:
                raise
        self._smbconf.create_set_share(name, value)

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._smbconf.share_names())

    def import_smbconf(
        self, src: ConfigStore, batch_size: typing.Optional[int] = 100
    ) -> None:
        """Import content from one SMBConf configuration object into the
        current SMBConf configuration object.

        Set batch_size to the maximum number of "shares" to import in one
        transaction. Set batch_size to None to use only one transaction.
        """
        if not self.writeable:
            raise ValueError("SMBConf is not writable")
        if batch_size is None:
            return self._import_smbconf_all(src)
        return self._import_smbconf_batched(src, batch_size)

    def _import_smbconf_all(self, src: ConfigStore) -> None:
        with self:
            for sname in src:
                self[sname] = src[sname]

    def _import_smbconf_batched(
        self, src: ConfigStore, batch_size: int
    ) -> None:
        # based on a comment in samba's source code for the net command
        # only import N 'shares' at a time so that the transaction does
        # not exceed talloc memory limits
        def _batch_keyfunc(item: tuple[int, str]) -> int:
            return item[0] // batch_size

        for _, snames in itertools.groupby(enumerate(src), _batch_keyfunc):
            with self:
                for _, sname in snames:
                    self[sname] = src[sname]



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import annotations

import json
import subprocess
import typing

from sambacc import samba_cmds

EXTERNAL: str = "external"
INTERNAL: str = "internal"


class HostState:
    T = typing.TypeVar("T", bound="HostState")

    def __init__(
        self, ref: str = "", items: typing.Optional[list[HostInfo]] = None
    ) -> None:
        self.ref: str = ref
        self.items: list[HostInfo] = items or []

    @classmethod
    def from_dict(cls: typing.Type[T], d: dict[str, typing.Any]) -> T:
        return cls(
            ref=d["ref"],
            items=[HostInfo.from_dict(i) for i in d.get("items", [])],
        )

    def __eq__(self, other: typing.Any) -> bool:
        return (
            self.ref == other.ref
            and len(self.items) == len(other.items)
            and all(s == o for (s, o) in zip(self.items, other.items))
        )


class HostInfo:
    T = typing.TypeVar("T", bound="HostInfo")

    def __init__(
        self, name: str = "", ipv4_addr: str = "", target: str = ""
    ) -> None:
        self.name = name
        self.ipv4_addr = ipv4_addr
        self.target = target

    @classmethod
    def from_dict(cls: typing.Type[T], d: dict[str, typing.Any]) -> T:
        return cls(
            name=d["name"],
            ipv4_addr=d["ipv4"],
            target=d.get("target", ""),
        )

    def __eq__(self, other: typing.Any) -> bool:
        return (
            self.name == other.name
            and self.ipv4_addr == other.ipv4_addr
            and self.target == other.target
        )


def parse(fh: typing.IO) -> HostState:
    return HostState.from_dict(json.load(fh))


def parse_file(path: str) -> HostState:
    with open(path) as fh:
        return parse(fh)


def match_target(state: HostState, target_name: str) -> list[HostInfo]:
    return [h for h in state.items if h.target == target_name]


def register(
    domain: str,
    hs: HostState,
    prefix: typing.Optional[list[str]] = None,
    target_name: str = EXTERNAL,
) -> bool:
    updated = False
    for item in match_target(hs, target_name):
        ip = item.ipv4_addr
        fqdn = "{}.{}".format(item.name, domain)
        cmd = samba_cmds.net["ads", "-P", "dns", "register", fqdn, ip]
        if prefix is not None:
            cmd.cmd_prefix = prefix
        subprocess.check_call(list(cmd))
        updated = True
    return updated


def parse_and_update(
    domain: str,
    source: str,
    previous: typing.Optional[HostState] = None,
    target_name: str = EXTERNAL,
    reg_func: typing.Callable = register,
) -> typing.Tuple[HostState, bool]:
    hs = parse_file(source)
    if previous is not None and hs == previous:
        # no changes
        return hs, False
    updated = reg_func(domain, hs, target_name=target_name)
    return hs, updated


# TODO: replace this with the common version added to simple_waiter
def watch(
    domain: str,
    source: str,
    update_func: typing.Callable,
    pause_func: typing.Callable,
    print_func: typing.Optional[typing.Callable],
) -> None:
    previous = None
    while True:
        try:
            previous, updated = update_func(domain, source, previous)
        except FileNotFoundError:
            if print_func:
                print_func(f"Source file [{source}] not found")
            updated = False
            previous = None
        if updated and print_func:
            print_func("Updating external dns registrations")
        try:
            pause_func()
        except KeyboardInterrupt:
            return



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import time
import typing


def generate_sleeps() -> typing.Iterator[int]:
    """Generate sleep times starting with short sleeps and then
    getting longer. This assumes that resources may take a bit of
    time to settle, but eventually reach a steadier state and don't
    require being checked as often.
    """
    total = 0
    while True:
        if total > 120:
            val = 60
        elif total > 10:
            val = 5
        else:
            val = 1
        yield val
        total += val


# It's a bit overkill to have a class for this but I didn't like messing
# around with functools.partial or functions returning functions for this.
# It's also nice to replace the sleep function for unit tests.
class Sleeper:
    """It waits only by sleeping. Nothing fancy."""

    def __init__(
        self, times: typing.Optional[typing.Iterator[int]] = None
    ) -> None:
        if times is None:
            times = generate_sleeps()
        self._times = times
        self._sleep = time.sleep

    def wait(self) -> None:
        self._sleep(next(self._times))

    def acted(self) -> None:
        """Inform the sleeper the caller reacted to a change and
        the sleeps should be reset.
        """
        self.times = generate_sleeps()


class Waiter(typing.Protocol):
    """Waiter protocol - interfaces common to all waiters."""

    def wait(self) -> None:
        """Pause execution for a time."""
        ...  # pragma: no cover

    def acted(self) -> None:
        """Inform that waiter that changes were made."""
        ...  # pragma: no cover


def watch(
    waiter: Waiter,
    initial_value: typing.Any,
    fetch_func: typing.Callable[..., typing.Any],
    compare_func: typing.Callable[..., typing.Tuple[typing.Any, bool]],
) -> None:
    """A very simple "event loop" that fetches current data with
    `fetch_func`, compares and updates state with `compare_func` and
    then waits for new events with `pause_func`.
    """
    previous = initial_value
    while True:
        try:
            previous, updated = compare_func(fetch_func(), previous)
        except FileNotFoundError:
            updated = False
            previous = None
        try:
            if updated:
                waiter.acted()
            waiter.wait()
        except KeyboardInterrupt:
            return



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import re
import subprocess
import typing

from sambacc import config
from sambacc import samba_cmds

_logger = logging.getLogger(__name__)


def provision(
    realm: str,
    dcname: str,
    admin_password: str,
    dns_backend: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    options: typing.Optional[typing.Iterable[tuple[str, str]]] = None,
) -> None:
    # this function is a direct translation of a previous shell script
    # as samba-tool is based on python libs, this function could possibly
    # be converted to import samba's libs and use that.
    _logger.info(f"Provisioning AD domain: realm={realm}")
    subprocess.check_call(
        _provision_cmd(
            realm,
            dcname,
            admin_password=admin_password,
            dns_backend=dns_backend,
            domain=domain,
            options=options,
        )
    )
    return


def join(
    realm: str,
    dcname: str,
    admin_password: str,
    dns_backend: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    options: typing.Optional[typing.Iterable[tuple[str, str]]] = None,
) -> None:
    _logger.info(f"Joining AD domain: realm={realm}")
    subprocess.check_call(
        _join_cmd(
            realm,
            dcname,
            admin_password=admin_password,
            dns_backend=dns_backend,
            options=options,
        )
    )


def create_user(
    name: str,
    password: str,
    surname: typing.Optional[str],
    given_name: typing.Optional[str],
    ou: typing.Optional[str] = None,
) -> None:
    cmd = _user_create_cmd(name, password, surname, given_name, ou)
    _logger.info("Creating user: %r", name)
    subprocess.check_call(cmd)


def create_group(name: str, ou: typing.Optional[str] = None) -> None:
    cmd = _group_add_cmd(name, ou)
    _logger.info("Creating group: %r", name)
    subprocess.check_call(cmd)


def create_ou(name: str) -> None:
    cmd = _ou_add_cmd(name)
    _logger.info("Creating organizational unit: %r", name)
    subprocess.check_call(cmd)


def add_group_members(group_name: str, members: list[str]) -> None:
    cmd = _group_add_members_cmd(group_name, members)
    _logger.info("Adding group members: %r", cmd)
    subprocess.check_call(cmd)


def _filter_opts(
    options: typing.Optional[typing.Iterable[tuple[str, str]]]
) -> list[tuple[str, str]]:
    _skip_keys = ["netbios name"]
    options = options or []
    return [(k, v) for (k, v) in options if k not in _skip_keys]


def _provision_cmd(
    realm: str,
    dcname: str,
    admin_password: str,
    dns_backend: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    options: typing.Optional[typing.Iterable[tuple[str, str]]] = None,
) -> list[str]:
    if not dns_backend:
        dns_backend = "SAMBA_INTERNAL"
    if not domain:
        domain = realm.split(".")[0].upper()
    cmd = samba_cmds.sambatool[
        "domain",
        "provision",
        f"--option=netbios name={dcname}",
        "--use-rfc2307",
        f"--dns-backend={dns_backend}",
        "--server-role=dc",
        f"--realm={realm}",
        f"--domain={domain}",
        f"--adminpass={admin_password}",
    ]
    cmd = cmd[
        [f"--option={okey}={oval}" for okey, oval in _filter_opts(options)]
    ]
    return cmd.argv()


def _join_cmd(
    realm: str,
    dcname: str,
    admin_password: str,
    dns_backend: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    options: typing.Optional[typing.Iterable[tuple[str, str]]] = None,
) -> list[str]:
    if not dns_backend:
        dns_backend = "SAMBA_INTERNAL"
    if not domain:
        domain = realm.split(".")[0].upper()
    cmd = samba_cmds.sambatool[
        "domain",
        "join",
        realm,
        "DC",
        f"-U{domain}\\Administrator",
        f"--option=netbios name={dcname}",
        f"--dns-backend={dns_backend}",
        f"--password={admin_password}",
    ]
    cmd = cmd[
        [f"--option={okey}={oval}" for okey, oval in _filter_opts(options)]
    ]
    return cmd.argv()


def _user_create_cmd(
    name: str,
    password: str,
    surname: typing.Optional[str],
    given_name: typing.Optional[str],
    ou: typing.Optional[str],
) -> list[str]:
    cmd = samba_cmds.sambatool[
        "user",
        "create",
        name,
        password,
    ].argv()
    if surname:
        cmd.append(f"--surname={surname}")
    if given_name:
        cmd.append(f"--given-name={given_name}")
    if ou:
        cmd.append(f"--userou=OU={ou}")
    return cmd


def _group_add_cmd(name: str, ou: typing.Optional[str]) -> list[str]:
    cmd = samba_cmds.sambatool[
        "group",
        "add",
        name,
    ].argv()
    if ou:
        cmd.append(f"--groupou=OU={ou}")
    return cmd


def _ou_add_cmd(name: str) -> list[str]:
    cmd = samba_cmds.sambatool[
        "ou",
        "add",
        f"OU={name}",
    ].argv()
    return cmd


def _group_add_members_cmd(group_name: str, members: list[str]) -> list[str]:
    cmd = samba_cmds.sambatool[
        "group",
        "addmembers",
        group_name,
        ",".join(members),
    ].argv()
    return cmd


def _ifnames() -> list[str]:
    import socket

    return [iface for _, iface in socket.if_nameindex()]


def filtered_interfaces(
    ic: config.DCInterfaceConfig, ifnames: typing.Optional[list[str]] = None
) -> list[str]:
    _include = re.compile(ic.include_pattern or "^.*$")
    _exclude = re.compile(ic.exclude_pattern or "^$")
    if ifnames is None:
        ifnames = _ifnames()
    return [
        name
        for name in ifnames
        if (name == "lo")
        or (_include.match(name) and not _exclude.match(name))
    ]



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Utilities for working with JSON data stored in a file system file.
"""

import fcntl
import json
import os
import typing

from sambacc.typelets import ExcType, ExcValue, ExcTraceback, Self

OPEN_RO = os.O_RDONLY
OPEN_RW = os.O_CREAT | os.O_RDWR


def open(path: str, flags: int, mode: int = 0o644) -> typing.IO:
    """A wrapper around open to open JSON files for read or read/write.
    `flags` must be os.open type flags. Use `OPEN_RO` and `OPEN_RW` for
    convenience.
    """
    return os.fdopen(os.open(path, flags, mode), "r+")


def load(
    fh: typing.IO, default: typing.Optional[dict[str, typing.Any]] = None
) -> typing.Any:
    """Similar to json.load, but returns the `default` value if fh refers to an
    empty file. fh must be seekable."""
    if fh.read(4) == "":
        # probe it to see if its an empty file
        data = default
    else:
        fh.seek(0)
        data = json.load(fh)
    return data


def dump(data: typing.Any, fh: typing.IO) -> None:
    """Similar to json.dump, but truncates the file before writing in order
    to avoid appending data to the file. fh must be seekable.
    """
    fh.seek(0)
    fh.truncate(0)
    json.dump(data, fh)


def flock(fh: typing.IO) -> None:
    """A simple wrapper around flock."""
    fcntl.flock(fh.fileno(), fcntl.LOCK_EX)


class ClusterMetaJSONHandle:
    def __init__(self, fh: typing.IO) -> None:
        self._fh = fh

    def load(self) -> typing.Any:
        return load(self._fh, {})

    def dump(self, data: typing.Any) -> None:
        dump(data, self._fh)
        self._fh.flush()
        os.fsync(self._fh)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> None:
        self._fh.close()


class ClusterMetaJSONFile:
    def __init__(self, path: str) -> None:
        self.path = path

    def open(
        self, *, read: bool = True, write: bool = False, locked: bool = False
    ) -> ClusterMetaJSONHandle:
        if read and write:
            flags = OPEN_RW
        elif read:
            flags = OPEN_RO
        else:
            raise ValueError("write-only not supported")
        fh = open(self.path, flags)
        try:
            if locked:
                flock(fh)
        except Exception:
            fh.close()
            raise
        return ClusterMetaJSONHandle(fh)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os

import pytest

import sambacc.addc


def _fake_samba_tool(path):
    fake_samba_tool = path / "fake_samba_tool.sh"
    with open(fake_samba_tool, "w") as fh:
        fh.write("#!/bin/sh\n")
        fh.write(f"[ -e {path}/fail ] && exit 1\n")
        fh.write(f'echo "$@" > {path}/args.out\n')
        fh.write("exit 0")
    os.chmod(fake_samba_tool, 0o700)
    return fake_samba_tool


def test_provision(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.provision("FOOBAR.TEST", "quux", "h4ckm3")
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "--realm=FOOBAR.TEST" in result
    assert "--option=netbios name=quux" in result
    assert "--dns-backend=SAMBA_INTERNAL" in result

    sambacc.addc.provision(
        "BARFOO.TEST",
        "quux",
        "h4ckm3",
        options=[
            ("ldap server require strong auth", "no"),
            ("dns zone scavenging", "yes"),
            ("ldap machine suffix", "ou=Machines"),
            ("netbios name", "flipper"),
        ],
    )
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "--realm=BARFOO.TEST" in result
    assert "--option=netbios name=quux" in result
    assert "--dns-backend=SAMBA_INTERNAL" in result
    assert "--option=ldap server require strong auth=no" in result
    assert "--option=dns zone scavenging=yes" in result
    assert "--option=ldap machine suffix=ou=Machines" in result
    assert "--option=netbios name=flipper" not in result

    open(tmp_path / "fail", "w").close()
    with pytest.raises(Exception):
        sambacc.addc.provision("FOOBAR.TEST", "quux", "h4ckm3")


def test_join(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.join("FOOBAR.TEST", "quux", "h4ckm3")
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "FOOBAR.TEST" in result
    assert "--option=netbios name=quux" in result
    assert "--dns-backend=SAMBA_INTERNAL" in result

    sambacc.addc.join(
        "BARFOO.TEST",
        "quux",
        "h4ckm3",
        options=[
            ("ldap server require strong auth", "no"),
            ("dns zone scavenging", "yes"),
            ("ldap machine suffix", "ou=Machines"),
            ("netbios name", "flipper"),
        ],
    )
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "BARFOO.TEST" in result
    assert "--option=netbios name=quux" in result
    assert "--dns-backend=SAMBA_INTERNAL" in result
    assert "--option=ldap server require strong auth=no" in result
    assert "--option=dns zone scavenging=yes" in result
    assert "--option=ldap machine suffix=ou=Machines" in result
    assert "--option=netbios name=flipper" not in result


def test_create_user(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.create_user("fflintstone", "b3dr0ck", "Flintstone", "Fred")
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "user create fflintstone" in result
    assert "--surname=Flintstone" in result
    assert "--given-name=Fred" in result


def test_create_ou(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.create_ou("quarry_workers")
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "ou add OU=quarry_workers" in result


def test_create_group(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.create_group("quarry_workers")
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "group add quarry_workers" in result


def test_add_group_members(tmp_path, monkeypatch):
    monkeypatch.setattr(
        sambacc.samba_cmds, "_GLOBAL_PREFIX", [_fake_samba_tool(tmp_path)]
    )

    sambacc.addc.add_group_members(
        "quarry_workers", ["fflintstone", "brubble"]
    )
    with open(tmp_path / "args.out") as fh:
        result = fh.read()
    assert "group addmembers quarry_workers" in result
    assert "fflintstone,brubble" in result


@pytest.mark.parametrize(
    "cfg,ifaces,expected",
    [
        ({}, ["foo", "bar"], None),
        (
            {"include_pattern": "^eth.*$"},
            ["lo", "eth0", "eth1", "biff1"],
            ["lo", "eth0", "eth1"],
        ),
        (
            {"include_pattern": "^nope$"},
            ["lo", "eth0", "eth1", "biff1"],
            ["lo"],
        ),
        (
            {"include_pattern": "^biff[0-9]+$"},
            ["lo", "eth0", "eth1", "biff1"],
            ["lo", "biff1"],
        ),
        (
            {"exclude_pattern": "^docker[0-9]+$"},
            ["lo", "eno1", "eno2", "docker0"],
            ["lo", "eno1", "eno2"],
        ),
        (
            {"exclude_pattern": "^.*$"},
            ["lo", "eno1", "eno2", "docker0"],
            ["lo"],
        ),
        (
            {
                "include_pattern": "^[ed].*$",
                "exclude_pattern": "^docker[0-9]+$",
            },
            ["lo", "eno1", "eno2", "docker0"],
            ["lo", "eno1", "eno2"],
        ),
        (
            {
                "include_pattern": "^[ed].*$",
                "exclude_pattern": "^.*0$",
            },
            ["lo", "dx1f2", "docker0"],
            ["lo", "dx1f2"],
        ),
    ],
)
def test_filtered_interfaces(cfg, ifaces, expected):
    ic = sambacc.config.DCInterfaceConfig(cfg)
    if cfg:
        assert ic.configured
        assert sambacc.addc.filtered_interfaces(ic, ifaces) == expected
    else:
        assert not ic.configured



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io
import os
import unittest

import pytest

import sambacc.config
import sambacc.opener

config1 = """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["demo", "stuff"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "demo": {
          "options": {
              "path": "/mnt/demo"
          }
       },
       "stuff": {
          "options": {
              "path": "/mnt/stuff"
          }
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}
"""

config2 = """
{
  "samba-container-config": "v0",
  "configs": {
    "foobar": {
      "shares": [
        "share"
      ],
      "globals": ["global0"],
      "instance_name": "GANDOLPH",
      "permissions": {
          "method": "none"
      }
    }
  },
  "shares": {
    "share": {
      "options": {
        "path": "/share",
        "read only": "no",
        "valid users": "sambauser",
        "guest ok": "no",
        "force user": "root"
      }
    }
  },
  "globals": {
    "global0": {
      "options": {
        "workgroup": "SAMBA",
        "security": "user",
        "server min protocol": "SMB2",
        "load printers": "no",
        "printing": "bsd",
        "printcap name": "/dev/null",
        "disable spoolss": "yes",
        "guest ok": "no"
      }
    }
  },
  "users": {
      "all_entries": [
        {"name": "bob", "password": "notSoSafe"},
        {"name": "alice", "password": "123fakeStreet"},
        {"name": "carol", "nt_hash": "B784E584D34839235F6D88A5382C3821"}
      ]
  },
  "_extra_junk": 0
}
"""

config3 = """
{
  "samba-container-config": "v0",
  "configs": {
    "foobar": {
      "shares": [
        "share"
      ],
      "globals": ["global0"],
      "instance_name": "RANDOLPH"
    }
  },
  "shares": {
    "share": {
      "options": {
        "path": "/share",
        "read only": "no",
        "valid users": "sambauser",
        "guest ok": "no",
        "force user": "root"
      },
      "permissions": {
          "method": "none"
      }
    }
  },
  "globals": {
    "global0": {
      "options": {
        "workgroup": "SAMBA",
        "security": "user",
        "server min protocol": "SMB2",
        "load printers": "no",
        "printing": "bsd",
        "printcap name": "/dev/null",
        "disable spoolss": "yes",
        "guest ok": "no"
      }
    }
  },
  "users": {
      "all_entries": [
        {"name": "bob", "uid": 2000, "gid": 2000,
         "password": "notSoSafe"},
        {"name": "alice","uid": 2001, "gid": 2001,
         "password": "123fakeStreet"},
        {"name": "carol",
         "nt_hash": "B784E584D34839235F6D88A5382C3821"}
      ]
  },
  "groups": {
      "all_entries": [
        {"name": "bobs", "gid": 2000},
        {"name": "alii", "gid": 2001}
      ]
  }
}
"""

ctdb_config1 = """
{
  "samba-container-config": "v0",
  "configs": {
    "ctdb1": {
      "shares": [
        "demo"
      ],
      "globals": [
        "global0"
      ],
      "instance_features": ["ctdb"],
      "instance_name": "ceeteedeebee"
    }
  },
  "shares": {
    "demo": {
      "options": {
        "path": "/share"
      }
    }
  },
  "globals": {
    "global0": {
      "options": {
        "security": "user",
        "load printers": "no",
        "printing": "bsd",
        "printcap name": "/dev/null",
        "disable spoolss": "yes",
        "guest ok": "no"
      }
    }
  },
  "users": {
    "all_entries": [
      {
        "name": "bob",
        "password": "notSoSafe"
      }
    ]
  }
}
"""

addc_config1 = """
{
  "samba-container-config": "v0",
  "configs": {
    "demo": {
      "instance_features": ["addc"],
      "domain_settings": "sink",
      "instance_name": "dc1"
    }
  },
  "domain_settings": {
    "sink": {
      "realm": "DOMAIN1.SINK.TEST",
      "short_domain": "DOMAIN1",
      "admin_password": "Passw0rd"
    }
  },
  "domain_groups": {
    "sink": [
      {"name": "friends"},
      {"name": "gothamites"}
    ]
  },
  "domain_users": {
    "sink": [
      {
        "name": "bwayne",
        "password": "1115Rose.",
        "given_name": "Bruce",
        "surname": "Wayne",
        "member_of": ["friends", "gothamites"]
      },
      {
        "name": "ckent",
        "password": "1115Rose.",
        "given_name": "Clark",
        "surname": "Kent",
        "member_of": ["friends"]
      }
    ]
  }
}
"""

addc_config2 = """
{
  "samba-container-config": "v0",
  "configs": {
    "demo": {
      "instance_features": ["addc"],
      "domain_settings": "sink",
      "instance_name": "dc1"
    }
  },
  "domain_settings": {
    "sink": {
      "realm": "DOMAIN1.SINK.TEST",
      "short_domain": "DOMAIN1",
      "admin_password": "Passw0rd"
    }
  },
  "organizational_units": {
    "sink": [
      {"name": "friends"}
    ]
  },
  "domain_groups": {
    "sink": [
      {
        "name": "friends",
        "ou": "friends"
      },
      {"name": "gothamites"}
    ]
  },
  "domain_users": {
    "sink": [
      {
        "name": "bwayne",
        "password": "1115Rose.",
        "given_name": "Bruce",
        "surname": "Wayne",
        "member_of": ["friends", "gothamites"],
        "ou": "friends"
      },
      {
        "name": "ckent",
        "password": "1115Rose.",
        "given_name": "Clark",
        "surname": "Kent",
        "member_of": ["friends"],
        "ou": "friends"
      }
    ]
  }
}
"""


class TestConfig(unittest.TestCase):
    def test_non_json(self):
        with self.assertRaises(Exception):
            with open(os.devnull) as fh:
                sambacc.config.GlobalConfig(fh)

    def test_empty_json(self):
        with self.assertRaises(ValueError):
            fh = io.StringIO("{}")
            sambacc.config.GlobalConfig(fh)

    def test_bad_version(self):
        with self.assertRaises(ValueError):
            fh = io.StringIO('{"samba-container-config":"foo"}')
            sambacc.config.GlobalConfig(fh)

    def test_valid_parse(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        assert isinstance(g, sambacc.config.GlobalConfig)

    def test_get_config(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        assert len(list(ic.shares())) == 2

    def test_fail_get_config(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        with self.assertRaises(KeyError):
            g.get("wobble")

    def test_get_global_opts(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        gopts = list(ic.global_options())
        assert ("server min protocol", "SMB2") in gopts

    def test_get_share_opts(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        shares = list(ic.shares())
        for share in shares:
            if share.name == "demo":
                assert ("path", "/mnt/demo") in list(share.share_options())
            elif share.name == "stuff":
                assert ("path", "/mnt/stuff") in list(share.share_options())
            else:
                raise AssertionError(share.name)

    def test_get_share_paths(self):
        fh = io.StringIO(config1)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        shares = list(ic.shares())
        for share in shares:
            if share.name == "demo":
                assert share.path() == "/mnt/demo"
            elif share.name == "stuff":
                assert share.path() == "/mnt/stuff"
            else:
                raise AssertionError(share.name)

    def test_unique_name(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        assert ("netbios name", "GANDOLPH") in list(ic.global_options())

    def test_many_global_opts(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        assert len(list(ic.global_options())) == (8 + 1)

    def test_some_users(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        users = list(ic.users())
        assert len(users) == 3
        assert users[0].username == "bob"
        assert users[0].uid == 1000
        assert users[0].gid == 1000
        pwline = ":".join(users[2].passwd_fields())
        assert pwline == "carol:x:1002:1002::/invalid:/bin/false"

    def test_some_groups(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        groups = list(ic.groups())
        assert len(groups) == 3
        assert groups[0].groupname == "bob"
        assert groups[0].gid == 1000

    def test_invalid_user_entry(self):
        rec = {"name": "foo", "uid": "fred"}
        with pytest.raises(ValueError):
            sambacc.config.UserEntry(None, rec, 0)
        rec = {"name": "foo", "uid": 2200, "gid": "beep"}
        with pytest.raises(ValueError):
            sambacc.config.UserEntry(None, rec, 0)

    def test_invalid_group_entry(self):
        rec = {"name": "foo", "gid": "boop"}
        with pytest.raises(ValueError):
            sambacc.config.GroupEntry(None, rec, 0)

    def test_user_entry_fields(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        rec = {"name": "jim", "uid": 2200, "gid": 2200}
        ue = sambacc.config.UserEntry(ic, rec, 0)
        assert ue.uid == 2200
        assert ue.gid == 2200
        assert ue.plaintext_passwd == ""

        rec = {"name": "jim", "password": "letmein"}
        ue = sambacc.config.UserEntry(ic, rec, 10)
        assert ue.uid == 1010
        assert ue.gid == 1010
        assert ue.plaintext_passwd == "letmein"

        rec = {"name": "jim", "nt_hash": "544849536973494D504F535349424C45"}
        ue = sambacc.config.UserEntry(ic, rec, 10)
        assert ue.uid == 1010
        assert ue.gid == 1010
        assert ue.plaintext_passwd == ""
        assert ue.nt_passwd == b"THISisIMPOSSIBLE"

    def test_group_entry_fields(self):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")

        rec = {"name": "hackers", "gid": 2200}
        ue = sambacc.config.GroupEntry(ic, rec, 0)
        assert ue.gid == 2200

        rec = {"name": "hackers"}
        ue = sambacc.config.GroupEntry(ic, rec, 20)
        assert ue.gid == 1020

    def test_explicitly_defined_groups(self):
        fh = io.StringIO(config3)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        groups = list(ic.groups())
        assert len(groups) == 3
        assert groups[0].groupname == "bobs"
        assert groups[0].gid == 2000
        assert groups[1].groupname == "alii"
        assert groups[1].gid == 2001
        assert groups[2].groupname == "carol"
        assert groups[2].gid == 1002


def test_read_config_files(tmpdir):
    fname = tmpdir / "sample.json"
    with open(fname, "w") as fh:
        fh.write(config1)
    sambacc.config.read_config_files([fname])


def test_read_config_files_noexist(tmpdir):
    fake1 = tmpdir / "fake1"
    fake2 = tmpdir / "fake2"
    with pytest.raises(ValueError):
        sambacc.config.read_config_files([fake1, fake2])


def test_read_config_files_realerr(tmpdir):
    if os.getuid() == 0:
        pytest.skip("test invalid when uid=0")
    fname = tmpdir / "sample.json"
    with open(fname, "w") as fh:
        fh.write(config1)
    # Prevent reading of the file to test any other error than
    # ENOENT is raised.
    os.chmod(fname, 0o333)
    try:
        with pytest.raises(OSError):
            sambacc.config.read_config_files([fname])
    finally:
        os.unlink(fname)


def test_tesd_config_files_realerr_rootok(monkeypatch):
    def err_open(*args):
        raise OSError("test!")

    monkeypatch.setattr(sambacc.opener.FileOpener, "open", err_open)
    fname = "/etc/foobar"
    with pytest.raises(OSError):
        sambacc.config.read_config_files([fname])


def test_instance_with_ctdb():
    c1 = sambacc.config.GlobalConfig(io.StringIO(config1))
    i1 = c1.get("foobar")
    assert not i1.with_ctdb

    c2 = sambacc.config.GlobalConfig(io.StringIO(ctdb_config1))
    i2 = c2.get("ctdb1")
    assert i2.with_ctdb


def test_instance_ctdb_smb_config():
    c1 = sambacc.config.GlobalConfig(io.StringIO(config1))
    i1 = c1.get("foobar")
    with pytest.raises(ValueError):
        i1.ctdb_smb_config()

    c2 = sambacc.config.GlobalConfig(io.StringIO(ctdb_config1))
    i2 = c2.get("ctdb1")
    csc = i2.ctdb_smb_config()
    gopts = dict(csc.global_options())
    assert gopts["clustering"] == "yes"
    assert gopts["include"] == "registry"
    assert csc.shares() == []


def test_instance_ctdb_config():
    c1 = sambacc.config.GlobalConfig(io.StringIO(config1))
    i1 = c1.get("foobar")
    assert i1.ctdb_config() == {}

    c2 = sambacc.config.GlobalConfig(io.StringIO(ctdb_config1))
    i2 = c2.get("ctdb1")
    cfg = i2.ctdb_config()
    assert "cluster_meta_uri" in cfg
    assert "nodes_path" in cfg
    assert "log_level" in cfg


def test_ad_dc_config_demo():
    c1 = sambacc.config.GlobalConfig(io.StringIO(addc_config1))
    i1 = c1.get("demo")
    assert i1.with_addc

    domcfg = i1.domain()
    assert domcfg.realm == "DOMAIN1.SINK.TEST"
    assert domcfg.short_domain == "DOMAIN1"
    assert domcfg.dcname == "dc1"

    dgroups = sorted(i1.domain_groups(), key=lambda v: v.groupname)
    assert len(dgroups) == 2
    assert dgroups[0].groupname == "friends"

    dusers = sorted(i1.domain_users(), key=lambda v: v.username)
    assert len(dusers) == 2
    assert dusers[0].username == "bwayne"


def test_ad_dc_ou_config_demo():
    c1 = sambacc.config.GlobalConfig(io.StringIO(addc_config2))
    i1 = c1.get("demo")
    assert i1.with_addc

    domou = sorted(i1.organizational_units(), key=lambda v: v.ou_name)
    assert len(domou) == 1
    assert domou[0].ou_name == "friends"

    dgroups = sorted(i1.domain_groups(), key=lambda v: v.groupname)
    assert len(dgroups) == 2
    assert dgroups[0].ou == "friends"

    dusers = sorted(i1.domain_users(), key=lambda v: v.username)
    assert len(dusers) == 2
    assert dusers[0].ou == "friends"


def test_ad_dc_invalid():
    c1 = sambacc.config.GlobalConfig(io.StringIO(config1))
    i1 = c1.get("foobar")
    assert not i1.with_addc

    with pytest.raises(ValueError):
        i1.domain()

    with pytest.raises(ValueError):
        list(i1.domain_users())

    with pytest.raises(ValueError):
        list(i1.domain_groups())

    with pytest.raises(ValueError):
        list(i1.organizational_units())


def test_share_config_no_path():
    j = """{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["flunk"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "flunk": {
          "options": {}
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}"""
    fh = io.StringIO(j)
    g = sambacc.config.GlobalConfig(fh)
    ic = g.get("foobar")
    shares = list(ic.shares())
    assert len(shares) == 1
    assert shares[0].path() is None


@pytest.mark.parametrize(
    "json_a,json_b,iname,expect_equal",
    [
        (config1, config1, "foobar", True),
        (addc_config1, addc_config1, "demo", True),
        (config1, config2, "foobar", False),
        (
            """{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["flunk"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "flunk": {
          "options": {"path": "/mnt/yikes"}
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}""",
            """{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["flunk"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "flunk": {
          "options": {"path": "/mnt/psych"}
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}""",
            "foobar",
            False,
        ),
        (
            """{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["flunk"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "flunk": {
          "options": {"path": "/mnt/yikes"}
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}""",
            """{
    "samba-container-config": "v0",
    "configs": {
        "foobar":{
          "shares": ["flunk"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "flunk": {
          "options": {"path": "/mnt/yikes"}
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB1"
           }
        }
    }
}""",
            "foobar",
            False,
        ),
    ],
    # setting ID to a numeric range makes it a lot easier to read the
    # output on the console, versus having pytest plop two large json
    # blobs for each "row" of inputs
    ids=iter(range(100)),
)
def test_instance_config_equality(json_a, json_b, iname, expect_equal):
    gca = sambacc.config.GlobalConfig(io.StringIO(json_a))
    gcb = sambacc.config.GlobalConfig(io.StringIO(json_b))
    instance_a = gca.get(iname)
    instance_b = gcb.get(iname)
    if expect_equal:
        assert instance_a == instance_b
    else:
        assert instance_a != instance_b


def test_permissions_config_default():
    c1 = sambacc.config.GlobalConfig(io.StringIO(config1))
    ic = c1.get("foobar")
    for share in ic.shares():
        assert share.permissions_config().method == "none"


def test_permissions_config_instance():
    c2 = sambacc.config.GlobalConfig(io.StringIO(config2))
    ic = c2.get("foobar")
    # TODO: improve test to ensure this isn't getting the default.  it does
    # work as designed based on coverage, but we shouldn't rely on that
    for share in ic.shares():
        assert share.permissions_config().method == "none"


def test_permissions_config_share():
    c3 = sambacc.config.GlobalConfig(io.StringIO(config3))
    ic = c3.get("foobar")
    # TODO: improve test to ensure this isn't getting the default.  it does
    # work as designed based on coverage, but we shouldn't rely on that
    for share in ic.shares():
        assert share.permissions_config().method == "none"


def test_permissions_config_options():
    pc = sambacc.config.PermissionsConfig(
        {
            "method": "initialize-share-perms",
            "status_xattr": "user.fake-stuff",
            "mode": "0777",
            "friendship": "always",
        }
    )
    opts = pc.options
    assert len(opts) == 2
    assert "mode" in opts
    assert "friendship" in opts


def _can_import_toml():
    """Return true if one valid toml module can be imported.
    Work around importorskip only supporting one module name.
    """
    try:
        __import__("tomllib")
        return True
    except ImportError:
        pass
    try:
        __import__("tomli")
        return True
    except ImportError:
        pass
    return False


@pytest.mark.parametrize(
    "json_str,ok",
    [
        pytest.param("{}", False, id="empty-json"),
        pytest.param('{"samba-container-config":"v0"}', True, id="minimal"),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "shares": [
                "foobar"
            ]
        }
    },
    "shares": {
        "foobar": {
            "options": {
                "a": "b"
            }
        }
    }
}
        """,
            True,
            id="one-share",
        ),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "shares": [
                "foobar"
            ]
        }
    },
    "shares": {
        "foobar": {
            "options": {
                "a": "b"
            }
        }
    },
    "chairs": {"maple": true}
}
        """,
            False,
            id="invalid-section",
        ),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "shares": [
                "foobar"
            ]
        }
    },
    "shares": {
        "foobar": {
            "options": {
                "a": "b"
            }
        }
    },
    "_chairs": {"maple": true}
}
        """,
            True,
            id="underscore-prefix",
        ),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "lazlo": "quux",
            "shares": [
                "foobar"
            ]
        }
    },
    "shares": {
        "foobar": {
            "options": {
                "a": "b"
            }
        }
    }
}
        """,
            False,
            id="bad-config-param",
        ),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "shares": "foobar"
        }
    },
    "shares": {
        "foobar": {
            "options": {
                "a": "b"
            }
        }
    }
}
        """,
            False,
            id="bad-config-type",
        ),
        pytest.param(
            """
{
    "samba-container-config": "v0",
    "configs": {
        "foobar": {
            "shares": [
                "foobar"
            ]
        }
    },
    "shares": {
        "foobar": {
            "blooper": true,
            "options": {
                "a": "b"
            }
        }
    }
}
        """,
            False,
            id="bad-share-prop",
        ),
    ],
)
def test_jsonschema_validation(json_str, ok):
    jsonschema = pytest.importorskip("jsonschema")

    cfg = sambacc.config.GlobalConfig()
    if ok:
        cfg.load(io.StringIO(json_str), require_validation=True)
    else:
        with pytest.raises((ValueError, jsonschema.ValidationError)):
            cfg.load(io.StringIO(json_str), require_validation=True)


@pytest.mark.parametrize(
    "toml_str,ok",
    [
        pytest.param("", False, id="empty"),
        pytest.param("#####FOO", False, id="just-a-comment"),
        pytest.param(
            """
samba-container-config = "v0"
""",
            True,
            id="minimal",
        ),
        pytest.param(
            """
samba-container-config = "v0"

# Define configurations
[configs.foobar]
shares = ["foobar"]

# Define share options
[shares.foobar.options]
a = "b"
""",
            True,
            id="one-share",
        ),
    ],
)
@pytest.mark.skipif(not _can_import_toml(), reason="no toml module")
def test_toml_configs_no_validation(toml_str, ok):
    cfg = sambacc.config.GlobalConfig()
    fh = io.BytesIO(toml_str.encode("utf8"))
    if ok:
        cfg.load(
            fh,
            require_validation=False,
            config_format=sambacc.config.ConfigFormat.TOML,
        )
    else:
        with pytest.raises(ValueError):
            cfg.load(
                fh,
                require_validation=False,
                config_format=sambacc.config.ConfigFormat.TOML,
            )


@pytest.mark.parametrize(
    "toml_str,ok",
    [
        pytest.param("", False, id="empty"),
        pytest.param("#####FOO", False, id="just-a-comment"),
        pytest.param(
            """
samba-container-config = "v0"
""",
            True,
            id="minimal",
        ),
        pytest.param(
            """
samba-container-config = "v0"

# Define configurations
[configs.foobar]
shares = ["foobar"]

# Define share options
[shares.foobar.options]
a = "b"
""",
            True,
            id="one-share",
        ),
        pytest.param(
            """
samba-container-config = "v0"

# Define configurations
[configs.foobar]
shares = ["foobar"]
instance_features = "Kibbe"

# Define share options
[shares.foobar.options]
a = "b"
""",
            False,
            id="bad-instance_features",
        ),
        pytest.param(
            """
samba-container-config = "v0"

# Define configurations
[configs.foobar]
shares = ["foobar"]
instance_features = ["ctdb"]

# Define share options
[shares.foobar.options]
a = "b"
""",
            True,
            id="ok-instance_features",
        ),
        pytest.param(
            """
samba-container-config = "v0"

[configs.demo]
shares = ["share"]
globals = ["default"]
instance_features = ["ctdb"]
instance_name = "SAMBA"

[shares.share.options]
"path" = "/share"
"read only" = "no"
"valid users" = "sambauser, otheruser"

[globals.default.options]
"security" = "user"
"server min protocol" = "SMB2"
"load printers" = "no"
"printing" = "bsd"
"printcap name" = "/dev/null"
"disable spoolss" = "yes"
"guest ok" = "no"

[[users.all_entries]]
name = "sambauser"
password = "samba"

[[users.all_entries]]
name = "otheruser"
password = "insecure321"
""",
            True,
            id="complex",
        ),
    ],
)
@pytest.mark.skipif(not _can_import_toml(), reason="no toml module")
def test_toml_configs_validation(toml_str, ok):
    jsonschema = pytest.importorskip("jsonschema")

    cfg = sambacc.config.GlobalConfig()
    fh = io.BytesIO(toml_str.encode("utf8"))
    if ok:
        cfg.load(
            fh,
            require_validation=True,
            config_format=sambacc.config.ConfigFormat.TOML,
        )
    else:
        with pytest.raises((ValueError, jsonschema.ValidationError)):
            cfg.load(
                fh,
                require_validation=True,
                config_format=sambacc.config.ConfigFormat.TOML,
            )


@pytest.mark.parametrize(
    "yaml_str,ok",
    [
        pytest.param("", False, id="empty"),
        pytest.param("#####FOO", False, id="just-a-comment"),
        pytest.param(
            """
samba-container-config: "v0"
""",
            True,
            id="minimal",
        ),
        pytest.param(
            """
samba-container-config: "v0"
# Define configurations
configs:
  foobar:
    shares:
      - foobar
shares:
  foobar:
    options:
      a: b
""",
            True,
            id="one-share",
        ),
    ],
)
def test_yaml_configs_no_validation(yaml_str, ok):
    pytest.importorskip("yaml")

    cfg = sambacc.config.GlobalConfig()
    fh = io.BytesIO(yaml_str.encode("utf8"))
    if ok:
        cfg.load(
            fh,
            require_validation=False,
            config_format=sambacc.config.ConfigFormat.YAML,
        )
    else:
        with pytest.raises(ValueError):
            cfg.load(
                fh,
                require_validation=False,
                config_format=sambacc.config.ConfigFormat.YAML,
            )


@pytest.mark.parametrize(
    "yaml_str,ok",
    [
        pytest.param("", False, id="empty"),
        pytest.param("#####FOO", False, id="just-a-comment"),
        pytest.param(
            """
samba-container-config: v0
""",
            True,
            id="minimal",
        ),
        pytest.param(
            """
samba-container-config: v0
# Define configurations
configs:
  foobar:
    shares:
      - foobar
# Define shares
shares:
  foobar:
    options:
      a: b
""",
            True,
            id="one-share",
        ),
        pytest.param(
            """
samba-container-config: v0
configs:
  foobar:
    instance_features: baroof
    shares:
      - foobar
shares:
  foobar:
    options:
      a: b
""",
            False,
            id="bad-instance_features",
        ),
        pytest.param(
            """
samba-container-config: v0
configs:
  foobar:
    instance_features:
      - ctdb
    shares:
      - foobar
shares:
  foobar:
    options:
      a: b
""",
            True,
            id="ok-instance_features",
        ),
        pytest.param(
            """
samba-container-config: "v0"
# Configure our demo instance
configs:
  demo:
    shares: ["share"]
    globals: ["default"]
    instance_features: ["ctdb"]
    instance_name: "SAMBA"
# Configure the share
shares:
  share:
    options:
      "path": "/share"
      "read only": "no"
      "valid users": "sambauser, otheruser"
# Configure globals
globals:
  default:
    options:
      "security": "user"
      "server min protocol": "SMB2"
      "load printers": "no"
      "printing": "bsd"
      "printcap name": "/dev/null"
      "disable spoolss": "yes"
      "guest ok": "no"
# Configure users
users:
  all_entries:
    - {"name": "sambauser", "password": "samba"}
    - {"name": "otheruser", "password": "insecure321"}
    - name: cooluser
      password: snarf
""",
            True,
            id="complex",
        ),
    ],
)
def test_yaml_configs_validation(yaml_str, ok):
    pytest.importorskip("yaml")
    jsonschema = pytest.importorskip("jsonschema")

    cfg = sambacc.config.GlobalConfig()
    fh = io.BytesIO(yaml_str.encode("utf8"))
    if ok:
        cfg.load(
            fh,
            require_validation=True,
            config_format=sambacc.config.ConfigFormat.YAML,
        )
    else:
        with pytest.raises((ValueError, jsonschema.ValidationError)):
            cfg.load(
                fh,
                require_validation=True,
                config_format=sambacc.config.ConfigFormat.YAML,
            )



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sys
import typing

import sambacc.join as joinutil

from .cli import (
    Context,
    Fail,
    Parser,
    best_waiter,
    commands,
    toggle_option,
)


def _print_join_error(err: typing.Any) -> None:
    print(f"ERROR: {err}", file=sys.stderr)
    for suberr in getattr(err, "errors", []):
        print(f"  - {suberr}", file=sys.stderr)


def _add_join_sources(joiner: joinutil.Joiner, cli: typing.Any) -> None:
    if cli.insecure or getattr(cli, "insecure_auto_join", False):
        upass = joinutil.UserPass(cli.username, cli.password)
        joiner.add_source(joinutil.JoinBy.PASSWORD, upass)
    if cli.files:
        for path in cli.join_files or []:
            joiner.add_source(joinutil.JoinBy.FILE, path)
    if cli.interactive:
        upass = joinutil.UserPass(cli.username)
        joiner.add_source(joinutil.JoinBy.INTERACTIVE, upass)


def _join_args(parser: Parser) -> None:
    parser.set_defaults(insecure=False, files=True, interactive=True)
    toggle_option(
        parser,
        arg="--insecure",
        dest="insecure",
        helpfmt="{} taking user/password from CLI or environment.",
    )
    toggle_option(
        parser,
        arg="--files",
        dest="files",
        helpfmt="{} reading user/password from JSON files.",
    )
    toggle_option(
        parser,
        arg="--interactive",
        dest="interactive",
        helpfmt="{} interactive password prompt.",
    )
    parser.add_argument(
        "--join-file",
        "-j",
        dest="join_files",
        action="append",
        help="Path to file with user/password in JSON format.",
    )


@commands.command(name="join", arg_func=_join_args)
def join(ctx: Context) -> None:
    """Perform a domain join. The supported sources for join
    can be provided by supplying command line arguments.
    This includes an *insecure* mode that sources the password
    from the CLI or environment. Use this only on
    testing/non-production purposes.
    """
    # maybe in the future we'll have more secure methods
    joiner = joinutil.Joiner(ctx.cli.join_marker, opener=ctx.opener)
    _add_join_sources(joiner, ctx.cli)
    try:
        joiner.join()
    except joinutil.JoinError as err:
        _print_join_error(err)
        raise Fail("failed to join to a domain")


def _must_join_args(parser: Parser) -> None:
    parser.set_defaults(insecure=False, files=True, wait=True)
    toggle_option(
        parser,
        arg="--insecure",
        dest="insecure",
        helpfmt="{} taking user/password from CLI or environment.",
    )
    toggle_option(
        parser,
        arg="--files",
        dest="files",
        helpfmt="{} reading user/password from JSON files.",
    )
    toggle_option(
        parser,
        arg="--wait",
        dest="wait",
        helpfmt="{} waiting until a join is done.",
    )
    parser.add_argument(
        "--join-file",
        "-j",
        dest="join_files",
        action="append",
        help="Path to file with user/password in JSON format.",
    )


@commands.command(name="must-join", arg_func=_must_join_args)
def must_join(ctx: Context) -> None:
    """If possible, perform an unattended domain join. Otherwise,
    exit or block until a join has been perfmed by another process.
    """
    joiner = joinutil.Joiner(ctx.cli.join_marker, opener=ctx.opener)
    if joiner.did_join():
        print("already joined")
        return
    # Interactive join is not allowed on must-join
    setattr(ctx.cli, "interactive", False)
    _add_join_sources(joiner, ctx.cli)
    if ctx.cli.wait:
        waiter = best_waiter(ctx.cli.join_marker, max_timeout=120)
        joinutil.join_when_possible(
            joiner, waiter, error_handler=_print_join_error
        )
    else:
        try:
            joiner.join()
        except joinutil.JoinError as err:
            _print_join_error(err)
            raise Fail(
                "failed to join to a domain - waiting for join is disabled"
            )



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import typing

from sambacc import ctdb
from sambacc import paths
import sambacc.nsswitch_loader as nsswitch

from . import config  # noqa: F401
from . import users  # noqa: F401
from .cli import commands, perms_handler, setup_steps, Context


_logger = logging.getLogger(__name__)


@setup_steps.command("nsswitch")
def _import_nsswitch(ctx: Context) -> None:
    # should nsswitch validation/edit be conditional only on ads?
    nss = nsswitch.NameServiceSwitchLoader("/etc/nsswitch.conf")
    nss.read()
    if not nss.winbind_enabled():
        nss.ensure_winbind_enabled()
        nss.write()


@setup_steps.command("smb_ctdb")
def _smb_conf_for_ctdb(ctx: Context) -> None:
    if ctx.instance_config.with_ctdb and ctx.expects_ctdb:
        _logger.info("Enabling ctdb in samba config file")
        ctdb.ensure_smb_conf(ctx.instance_config)


@setup_steps.command("ctdb_config")
def _ctdb_conf_for_ctdb(ctx: Context) -> None:
    if ctx.instance_config.with_ctdb and ctx.expects_ctdb:
        _logger.info("Ensuring ctdb config")
        ctdb.ensure_ctdb_conf(ctx.instance_config)


@setup_steps.command("ctdb_nodes")
def _ctdb_nodes_exists(ctx: Context) -> None:
    if ctx.instance_config.with_ctdb and ctx.expects_ctdb:
        _logger.info("Ensuring ctdb nodes file")
        persistent_path = ctx.instance_config.ctdb_config()["nodes_path"]
        ctdb.ensure_ctdb_nodes(
            ctdb_nodes=ctdb.read_ctdb_nodes(persistent_path),
            real_path=persistent_path,
        )


@setup_steps.command("ctdb_etc")
def _ctdb_etc_files(ctx: Context) -> None:
    if ctx.instance_config.with_ctdb and ctx.expects_ctdb:
        _logger.info("Ensuring ctdb etc files")
        ctdb.ensure_ctdbd_etc_files(iconfig=ctx.instance_config)


@setup_steps.command("share_paths")
@commands.command(name="ensure-share-paths")
def ensure_share_paths(ctx: Context) -> None:
    """Ensure the paths defined by the configuration exist."""
    # currently this is completely ignorant of things like vfs
    # modules that might "virtualize" the share path. It just
    # assumes that the path in the configuration is an absolute
    # path in the file system.
    for share in ctx.instance_config.shares():
        path = share.path()
        if not path:
            continue
        _logger.info(f"Ensuring share path: {path}")
        paths.ensure_share_dirs(path)
        _logger.info(f"Updating permissions if needed: {path}")
        perms_handler(share.permissions_config(), path).update()


_default_setup_steps = [
    "config",
    "users",
    "smb_ctdb",
    "users_passdb",
    "nsswitch",
]


def setup_step_names():
    """Return a list of names for the steps that init supports."""
    return list(setup_steps.dict().keys())


@commands.command(name="init")
def init_container(
    ctx: Context, steps: typing.Optional[typing.Iterable[str]] = None
) -> None:
    """Initialize the entire container environment."""
    steps = _default_setup_steps if steps is None else list(steps)
    cmds = setup_steps.dict()
    for step_name in steps:
        cmds[step_name].cmd_func(ctx)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2024  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from unittest import mock

import pytest

from sambacc.commands import skips


@pytest.mark.parametrize(
    "value,rtype",
    [
        ("always:", skips.SkipAlways),
        ("file:/var/lib/womble", skips.SkipFile),
        ("file:!/var/lib/zomble", skips.SkipFile),
        ("env:LIMIT==none", skips.SkipEnv),
        ("env:LIMIT!=everybody", skips.SkipEnv),
        ("env:LIMIT=everybody", ValueError),
        ("env:LIMIT", ValueError),
        ("file:", ValueError),
        ("always:forever", ValueError),
        ("klunk:", KeyError),
    ],
)
def test_parse(value, rtype):
    if issubclass(rtype, BaseException):
        with pytest.raises(rtype):
            skips.parse(value)
        return
    skf = skips.parse(value)
    assert isinstance(skf, rtype)


@pytest.mark.parametrize(
    "value,ret",
    [
        ("file:/var/lib/foo/a", "skip-if-file-exists: /var/lib/foo/a exists"),
        (
            "file:!/var/lib/bar/a",
            "skip-if-file-missing: /var/lib/bar/a missing",
        ),
        ("file:/etc/blat", None),
        ("env:PLINK==0", "env var: PLINK -> 0 == 0"),
        ("env:PLINK!=88", "env var: PLINK -> 0 != 88"),
        ("env:PLONK==enabled", None),
        ("always:", "always skip"),
    ],
)
def test_method_test(value, ret, monkeypatch):
    def _exists(p):
        rv = p.startswith("/var/lib/foo/")
        return rv

    monkeypatch.setattr("os.path.exists", _exists)
    monkeypatch.setenv("PLINK", "0")
    monkeypatch.setenv("PLONK", "disabled")
    skf = skips.parse(value)
    ctx = mock.MagicMock()
    assert skf.test(ctx) == ret


def test_test(monkeypatch):
    def _exists(p):
        rv = p.startswith("/var/lib/foo/")
        return rv

    monkeypatch.setattr("os.path.exists", _exists)
    monkeypatch.setenv("PLINK", "0")
    monkeypatch.setenv("PLONK", "disabled")

    conds = [
        skips.SkipEnv("==", "PLINK", "1"),
        skips.SkipEnv("!=", "PLONK", "disabled"),
        skips.SkipAlways(),
    ]
    ctx = mock.MagicMock()
    assert skips.test(ctx, conditions=conds) == "always skip"
    conds = conds[:-1]
    assert not skips.test(ctx, conditions=conds)
    monkeypatch.setenv("PLINK", "1")
    assert skips.test(ctx, conditions=conds) == "env var: PLINK -> 1 == 1"

    ctx.cli.skip_conditions = conds
    assert skips.test(ctx) == "env var: PLINK -> 1 == 1"


def test_help_info():
    txt = skips._help_info()
    assert "file:" in txt
    assert "env:" in txt
    assert "always:" in txt


def test_parse_hack():
    import argparse

    with pytest.raises(argparse.ArgumentTypeError):
        skips.parse("?")



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import subprocess
import typing

from sambacc import config
from sambacc import samba_cmds


class LoaderError(Exception):
    pass


def template_config(
    fh: typing.IO, iconfig: config.SambaConfig, enc: typing.Callable = str
) -> None:
    fh.write(enc("[global]\n"))
    for gkey, gval in iconfig.global_options():
        fh.write(enc(f"\t{gkey} = {gval}\n"))

    for share in iconfig.shares():
        fh.write(enc("\n[{}]\n".format(share.name)))
        for skey, sval in share.share_options():
            fh.write(enc(f"\t{skey} = {sval}\n"))


class NetCmdLoader:
    _net_conf = samba_cmds.net["conf"]

    def _cmd(
        self, *args: str, **kwargs: typing.Any
    ) -> tuple[list[str], typing.Any]:
        cmd = list(self._net_conf[args])
        return cmd, subprocess.Popen(cmd, **kwargs)

    def _check(self, cli: typing.Any, proc: subprocess.Popen) -> None:
        ret = proc.wait()
        if ret != 0:
            raise LoaderError("failed to run {}".format(cli))

    def import_config(self, iconfig: config.InstanceConfig) -> None:
        """Import to entire instance config to samba config."""
        cli, proc = self._cmd("import", "/dev/stdin", stdin=subprocess.PIPE)
        template_config(proc.stdin, iconfig, enc=samba_cmds.encode)
        proc.stdin.close()
        self._check(cli, proc)

    def dump(self, out: typing.IO) -> None:
        """Dump the current smb config in an smb.conf format.
        Writes the dump to `out`.
        """
        cli, proc = self._cmd("list", stdout=out)
        self._check(cli, proc)

    def _parse_shares(self, fh: typing.IO) -> typing.Iterable[str]:
        out = []
        for line in fh.readlines():
            line = line.strip().decode("utf8")
            if line == "global":
                continue
            out.append(line)
        return out

    def current_shares(self) -> typing.Iterable[str]:
        """Returns a list of current shares."""
        cli, proc = self._cmd("listshares", stdout=subprocess.PIPE)
        # read and parse shares list
        try:
            shares = self._parse_shares(proc.stdout)
        finally:
            self._check(cli, proc)
        return shares

    def set(self, section: str, param: str, value: str) -> None:
        """Set an individual config parameter."""
        cli, proc = self._cmd("setparm", section, param, value)
        self._check(cli, proc)



#!/usr/bin/python3
# --- GENERATED FILE --- DO NOT EDIT --- #
# --- generated from: conf-v0.schema.yaml

SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "mailto:phlogistonjohn+sambacc-v0@asynchrono.us",
    "title": "sambacc configuration",
    "description": (
        "The configuration for the sambacc tool. sambacc configures Samba and"
        " the container\nenvironment to fit Samba's unique needs. This"
        " configuration can hold configuration\nfor more than one server"
        ' "instance". The "configs" section contains one or'
        " more\nconfiguration with a name that can be selected at runtime."
        " Share definitions\nand samba global configuration blocks can be"
        " mixed and matched.\n"
    ),
    "type": "object",
    "$defs": {
        "section_choices": {
            "description": (
                "Selects sub-sections from elsewhere in the configuration.\n"
            ),
            "type": "array",
            "items": {"type": "string"},
        },
        "feature_flags": {
            "description": (
                "Feature flags are used to enable specific, wide-ranging,"
                " features of\nsambacc. For example, it is used to enable"
                " clustered mode with ctdb.\n"
            ),
            "type": "array",
            "items": {"enum": ["addc", "ctdb"]},
        },
        "samba_options": {
            "description": (
                "A mapping of values that will be passed into the smb.conf"
                " (or equivalent)\nto directly configure Samba.\n"
            ),
            "type": "object",
            "additionalProperties": {"type": "string"},
        },
        "permissions_config": {
            "description": (
                "Settings that enable and manage sambacc's permissions"
                " management support.\n"
            ),
            "type": "object",
            "properties": {
                "method": {
                    "description": (
                        "Backend method for controlling permissions on shares"
                    ),
                    "type": "string",
                },
                "status_xattr": {
                    "description": (
                        "xattr name used to store permissions state"
                    ),
                    "type": "string",
                },
            },
            "additionalProperties": {"type": "string"},
        },
        "user_entry": {
            "description": (
                "A user that will be instantiated in the local container"
                " environment to\nin order to provide access to smb shares.\n"
            ),
            "type": "object",
            "properties": {
                "name": {"description": "The user's name", "type": "string"},
                "uid": {
                    "description": "The Unix UID the user should have",
                    "type": "integer",
                },
                "gid": {
                    "description": "The Unix GID the user should have",
                    "type": "integer",
                },
                "nt_hash": {
                    "description": "An NT-Hashed password",
                    "type": "string",
                },
                "password": {
                    "description": "A plain-text password",
                    "type": "string",
                },
            },
            "required": ["name"],
            "additionalProperties": False,
        },
        "group_entry": {
            "description": (
                "A group that will be instantiated in the local container"
                " environment to\nin order to provide access to smb shares.\n"
            ),
            "type": "object",
            "properties": {
                "name": {"description": "The group name", "type": "string"},
                "gid": {
                    "description": "The Unix GID the group should have",
                    "type": "integer",
                },
            },
            "required": ["name"],
            "additionalProperties": False,
        },
        "domain_user_entry": {
            "description": (
                "A user that will be created in the specified AD domain."
                " These\nusers are populated in the directory after the"
                " domain is provisioned.\n"
            ),
            "type": "object",
            "properties": {
                "name": {"description": "The user's name", "type": "string"},
                "surname": {
                    "description": "A surname for the user",
                    "type": "string",
                },
                "given_name": {
                    "description": "A given name for the user",
                    "type": "string",
                },
                "uid": {"type": "integer"},
                "gid": {"type": "integer"},
                "nt_hash": {"type": "string"},
                "password": {
                    "description": "A plain-text password",
                    "type": "string",
                },
                "ou": {
                    "description": (
                        "A organizational unit that the user should belong to"
                    ),
                    "type": "string",
                },
                "member_of": {
                    "description": (
                        "A list of group names that the user should belong to"
                    ),
                    "type": "array",
                    "items": {"type": "string"},
                },
            },
            "required": ["name"],
            "additionalProperties": False,
        },
        "domain_group_entry": {
            "description": (
                "A group that will be created in the specified AD domain."
                " These\ngroups are populated in the directory after the"
                " domain is provisioned.\n"
            ),
            "type": "object",
            "properties": {
                "name": {"description": "The group name", "type": "string"},
                "gid": {"type": "integer"},
                "ou": {
                    "description": (
                        "A organizational unit that the user should belong to"
                    ),
                    "type": "string",
                },
            },
            "required": ["name"],
            "additionalProperties": False,
        },
        "organizational_unit_entry": {
            "description": (
                "A organizational unit that will be created in the specified"
                " AD domain. These\ngroups are populated in the directory"
                " after the domain is provisioned.\n"
            ),
            "type": "object",
            "properties": {
                "name": {
                    "description": "The organizational unit name",
                    "type": "string",
                }
            },
            "required": ["name"],
            "additionalProperties": False,
        },
    },
    "properties": {
        "samba-container-config": {
            "type": "string",
            "title": "Configuration Format Version",
            "description": (
                "A short version string that assists in allowing the"
                " configuration\nformat to (some day) support incompatible"
                " version changes.\n(It is unique to the configuration and is"
                " not the version of sambacc)\n"
            ),
        },
        "configs": {
            "title": "Container Configurations",
            "description": (
                "A mapping of named configurations (instances) to top-level"
                " configuration\nblocks. A usable configuration file must"
                " have at least one configuration,\nbut more than one is"
                " supported.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "shares": {"$ref": "#/$defs/section_choices"},
                    "globals": {"$ref": "#/$defs/section_choices"},
                    "instance_features": {"$ref": "#/$defs/feature_flags"},
                    "permissions": {"$ref": "#/$defs/permissions_config"},
                    "instance_name": {
                        "description": (
                            "A name that will be set for the server"
                            " instance.\n"
                        ),
                        "type": "string",
                    },
                    "domain_settings": {
                        "description": (
                            "The name of the domain settings. Only used with"
                            " 'ADDC' feature flag.\n"
                        ),
                        "type": "string",
                    },
                },
                "additionalProperties": False,
            },
        },
        "shares": {
            "description": (
                "A mapping of share name to share specific configuration. A"
                ' share can\nhave "options" that are passed to Samba. Shares'
                ' can have an optional\n"permissions" section for managing'
                " permissions/acls in sambacc.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "options": {"$ref": "#/$defs/samba_options"},
                    "permissions": {"$ref": "#/$defs/permissions_config"},
                },
                "additionalProperties": False,
            },
        },
        "globals": {
            "description": (
                "A mapping of samba global configuration blocks. The global"
                " section names\nare not passed to Samba. All sections"
                " selected by a configuration are\nmerged together before"
                " passing to Samba.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {"options": {"$ref": "#/$defs/samba_options"}},
                "additionalProperties": False,
            },
        },
        "domain_settings": {
            "description": (
                "A mapping of AD DC domain configuration keys to domain"
                " configurations.\nThese parameters are used when"
                " provisioning an AD DC instance.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "realm": {"type": "string"},
                    "short_domain": {"type": "string"},
                    "admin_password": {"type": "string"},
                    "interfaces": {
                        "type": "object",
                        "properties": {
                            "include_pattern": {
                                "type": "string",
                                "description": (
                                    "A regular expression that must match for"
                                    " a network interface\nto be included in"
                                    " the AD DC interfaces list.\n"
                                ),
                            },
                            "exclude_pattern": {
                                "type": "string",
                                "description": (
                                    "A regular expression that must not match"
                                    " for a network interface\nto be included"
                                    " in the AD DC interfaces list.\n"
                                ),
                            },
                        },
                    },
                },
                "required": ["realm"],
                "additionalProperties": False,
            },
        },
        "users": {
            "description": (
                "Users to add to the container environment in order to"
                " provide\nShare access-control without becoming a domain"
                " member server.\n"
            ),
            "type": "object",
            "properties": {
                "all_entries": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/user_entry"},
                }
            },
        },
        "groups": {
            "description": (
                "Groups to add to the container environment in order to"
                " provide\nShare access-control without becoming a domain"
                " member server.\n"
            ),
            "type": "object",
            "properties": {
                "all_entries": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/group_entry"},
                }
            },
        },
        "domain_users": {
            "description": (
                "The domain_users section defines initial users that will be"
                " automatically\nadded to a newly provisioned domain. This"
                " section is a mapping of the\ndomain settings name to a list"
                " of domain user entries.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "array",
                "items": {"$ref": "#/$defs/domain_user_entry"},
            },
        },
        "domain_groups": {
            "description": (
                "The domain_groups section defines initial groups that will"
                " be\nautomatically added to a newly provisioned domain. This"
                " section is\na mapping of the domain settings name to a list"
                " of domain group entries.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "array",
                "items": {"$ref": "#/$defs/domain_group_entry"},
            },
        },
        "organizational_units": {
            "description": (
                "The organizational_unit section defines initial"
                " organizational unit that will be\nautomatically added to a"
                " newly provisioned domain. This section is\na mapping of the"
                " domain settings name to a list of domain group entries.\n"
            ),
            "type": "object",
            "additionalProperties": {
                "type": "array",
                "items": {"$ref": "#/$defs/organizational_unit_entry"},
            },
        },
        "ctdb": {
            "type": "object",
            "additionalProperties": {"type": "string"},
        },
    },
    "additionalProperties": False,
    "required": ["samba-container-config"],
    "patternProperties": {"^_": True},
}



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing

from sambacc import config

# Do the samba python bindings not export any useful constants?
ACB_DISABLED = 0x00000001
ACB_NORMAL = 0x00000010
ACB_PWNOEXP = 0x00000200


def _samba_modules() -> tuple[typing.Any, typing.Any]:
    from samba.samba3 import param  # type: ignore
    from samba.samba3 import passdb  # type: ignore

    return param, passdb


class PassDBLoader:
    def __init__(self, smbconf: typing.Any = None) -> None:
        param, passdb = _samba_modules()
        lp = param.get_context()
        if smbconf is None:
            lp.load_default()
        else:
            lp.load(smbconf)
        passdb.set_secrets_dir(lp.get("private dir"))
        self._pdb = passdb.PDB(lp.get("passdb backend"))
        self._passdb = passdb

    def add_user(self, user_entry: config.UserEntry) -> None:
        if not (user_entry.nt_passwd or user_entry.plaintext_passwd):
            raise ValueError(
                f"user entry {user_entry.username} lacks password value"
            )
        # probe for an existing user, by name
        try:
            samu = self._pdb.getsampwnam(user_entry.username)
        except self._passdb.error:
            samu = None
        # if it doesn't exist, create it
        if samu is None:
            # FIXME, research if there are better flag values to use
            acb = ACB_NORMAL | ACB_PWNOEXP
            self._pdb.create_user(user_entry.username, acb)
            samu = self._pdb.getsampwnam(user_entry.username)
        acb = samu.acct_ctrl
        # update password/metadata
        if user_entry.nt_passwd:
            samu.nt_passwd = user_entry.nt_passwd
        elif user_entry.plaintext_passwd:
            samu.plaintext_passwd = user_entry.plaintext_passwd
        # Try to mimic the behavior of smbpasswd and clear the account disabled
        # flag when adding or updating the user.
        # We don't expect granular, on the fly, user management in the
        # container, so it seems pointless to have a user that can't log in.
        if acb & ACB_DISABLED:
            samu.acct_ctrl = acb & ~ACB_DISABLED
        # update the db
        self._pdb.update_sam_account(samu)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import pytest

from sambacc import jfile


def test_open(tmpdir):
    with pytest.raises(FileNotFoundError):
        jfile.open(tmpdir / "a.json", jfile.OPEN_RO)
    fh = jfile.open(tmpdir / "a.json", jfile.OPEN_RW)
    assert fh is not None
    fh.close()


def test_laod(tmpdir):
    with jfile.open(tmpdir / "a.json", jfile.OPEN_RW) as fh:
        data = jfile.load(fh, ["missing"])
    assert data == ["missing"]

    with open(tmpdir / "a.json", "w") as fh:
        fh.write('{"present": true}\n')
    with jfile.open(tmpdir / "a.json", jfile.OPEN_RW) as fh:
        data = jfile.load(fh, ["missing"])
    assert data == {"present": True}


def test_dump(tmpdir):
    with jfile.open(tmpdir / "a.json", jfile.OPEN_RW) as fh:
        jfile.dump({"something": "good"}, fh)

    with jfile.open(tmpdir / "a.json", jfile.OPEN_RO) as fh:
        data = jfile.load(fh)
    assert data == {"something": "good"}

    with jfile.open(tmpdir / "a.json", jfile.OPEN_RW) as fh:
        jfile.dump({"something": "better"}, fh)

    with jfile.open(tmpdir / "a.json", jfile.OPEN_RO) as fh:
        data = jfile.load(fh)
    assert data == {"something": "better"}


def test_flock(tmpdir):
    import time
    import threading

    def sleepy_update(path):
        with jfile.open(path, jfile.OPEN_RW) as fh:
            jfile.flock(fh)
            data = jfile.load(fh, [0])
            time.sleep(0.2)
            data.append(data[-1] + 1)
            jfile.dump(data, fh)

    fpath = tmpdir / "a.json"
    t1 = threading.Thread(target=sleepy_update, args=(fpath,))
    t1.start()
    t2 = threading.Thread(target=sleepy_update, args=(fpath,))
    t2.start()
    t1.join()
    t2.join()

    with jfile.open(fpath, jfile.OPEN_RW) as fh:
        jfile.flock(fh)
        data = jfile.load(fh)
    assert data == [0, 1, 2]



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import errno
import http
import http.server
import os
import sys
import threading
import urllib.request

import pytest

import sambacc.url_opener


class _Server:
    def __init__(self, port=8111):
        port = int(os.environ.get("SAMBACC_TEST_HTTP_PORT", port))
        self._port = port
        self._server = http.server.HTTPServer(("127.0.0.1", port), _Handler)

    @property
    def port(self):
        return self._port

    def start(self):
        self._t = threading.Thread(target=self._server.serve_forever)
        self._t.start()

    def stop(self):
        sys.stdout.flush()
        self._server.shutdown()
        self._t.join()


class _Handler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        word = self.path.split("/")[-1]
        method = f"get_{word}"
        return getattr(self, method)()

    def get_a(self):
        return self._ok("Wilbur was Right")

    def get_b(self):
        return self._ok("This is a test")

    def get_err404(self):
        self._err(http.HTTPStatus.NOT_FOUND, "Not Found")

    def get_err401(self):
        self._err(http.HTTPStatus.UNAUTHORIZED, "Unauthorized")

    def get_err403(self):
        self._err(http.HTTPStatus.FORBIDDEN, "Forbidden")

    def _ok(self, value):
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-Type", "text/plain")
        self.send_header("Content-Length", str(len(value)))
        self.end_headers()
        self.wfile.write(value.encode("utf8"))

    def _err(self, err_value, err_msg):
        self.send_response(err_value)
        self.send_header("Content-Type", "text/plain")
        self.send_header("Content-Length", str(len(err_msg)))
        self.end_headers()
        self.wfile.write(err_msg.encode("utf8"))


@pytest.fixture(scope="module")
def http_server():
    srv = _Server()
    srv.start()
    try:
        yield srv
    finally:
        srv.stop()


def test_success_1(http_server):
    url = f"http://localhost:{http_server.port}/a"
    opener = sambacc.url_opener.URLOpener()
    res = opener.open(url)
    assert res.read() == b"Wilbur was Right"


def test_success_2(http_server):
    url = f"http://localhost:{http_server.port}/b"
    opener = sambacc.url_opener.URLOpener()
    res = opener.open(url)
    assert res.read() == b"This is a test"


def test_error_404(http_server):
    url = f"http://localhost:{http_server.port}/err404"
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(OSError) as err:
        opener.open(url)
    assert err.value.status == 404
    assert err.value.errno == errno.ENOENT


def test_error_401(http_server):
    url = f"http://localhost:{http_server.port}/err401"
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(OSError) as err:
        opener.open(url)
    assert err.value.status == 401
    assert err.value.errno == errno.EPERM


def test_error_403(http_server):
    url = f"http://localhost:{http_server.port}/err403"
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(OSError) as err:
        opener.open(url)
    assert err.value.status == 403
    # No errno mapped for this one


def test_map_errno(http_server):
    url = f"http://localhost:{http_server.port}/err401"
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(OSError) as err:
        opener.open(url)
    # do not replace an existing errno
    err.value.errno = errno.EIO
    sambacc.url_opener._map_errno(err.value)
    assert err.value.errno == errno.EIO


def test_unknown_url():
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(sambacc.url_opener.SchemeNotSupported):
        opener.open("bloop://foo/bar/baz")


def test_unknown_url_type():
    opener = sambacc.url_opener.URLOpener()
    with pytest.raises(sambacc.url_opener.SchemeNotSupported):
        opener.open("bonk-bonk-bonk")


def test_value_error_during_handling():
    class H(urllib.request.BaseHandler):
        def bonk_open(self, req):
            raise ValueError("fiddlesticks")

    class UO(sambacc.url_opener.URLOpener):
        _handlers = sambacc.url_opener.URLOpener._handlers + [H]

    opener = UO()
    with pytest.raises(ValueError) as err:
        opener.open("bonk:bonk")
    assert str(err.value) == "fiddlesticks"



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import annotations

import io
import json
import logging
import time
import typing
import urllib.request
import uuid

from . import url_opener
from .typelets import ExcType, ExcValue, ExcTraceback, Self

_RADOSModule = typing.Any
_RADOSObject = typing.Any

_CHUNK_SIZE = 4 * 1024

_logger = logging.getLogger(__name__)


class RADOSUnsupported(Exception):
    pass


class _RADOSInterface:
    api: _RADOSModule
    client_name: str
    full_name: bool

    def Rados(self) -> _RADOSObject:
        name = rados_id = ""
        if self.full_name:
            name = self.client_name
        else:
            rados_id = self.client_name
        _logger.debug("Creating RADOS connection")
        return self.api.Rados(
            name=name,
            rados_id=rados_id,
            conffile=self.api.Rados.DEFAULT_CONF_FILES,
        )


class _RADOSHandler(urllib.request.BaseHandler):
    _interface: typing.Optional[_RADOSInterface] = None

    def rados_open(self, req: urllib.request.Request) -> typing.IO:
        """Open a rados-style url. Called from urllib."""
        if self._interface is None:
            raise RADOSUnsupported()
        rinfo = parse_rados_uri(req)
        if rinfo.get("subtype") == "mon-config-key":
            return _get_mon_config_key(self._interface, rinfo["path"])
        return RADOSObjectRef(
            self._interface, rinfo["pool"], rinfo["ns"], rinfo["key"]
        )

    def get_object(
        self, uri: str, *, must_exist: bool = False
    ) -> RADOSObjectRef:
        """Return a rados object reference for the given rados uri. The uri
        must refer to a rados object only as the RADOSObjectRef can do various
        rados-y things, more than an IO requires.
        """
        if self._interface is None:
            raise RADOSUnsupported()
        rinfo = parse_rados_uri(urllib.request.Request(uri))
        if rinfo.get("type") != "rados":
            raise ValueError("only rados URI values supported")
        if rinfo.get("subtype") == "mon-config-key":
            raise ValueError("only rados object URI values supported")
        return RADOSObjectRef(
            self._interface,
            rinfo["pool"],
            rinfo["ns"],
            rinfo["key"],
            must_exist=must_exist,
        )


# it's quite annoying to have a read-only typing.IO we're forced to
# have so many stub methods. Go's much more granular io interfaces for
# readers/writers is much nicer for this.
class RADOSObjectRef(typing.IO):
    def __init__(
        self,
        interface: _RADOSInterface,
        pool: str,
        ns: str,
        key: str,
        *,
        must_exist: bool = True,
    ) -> None:
        self._pool = pool
        self._ns = ns
        self._key = key
        self._lock_description = "sambacc RADOS library"
        self._lock_duration = None

        self._open(interface)
        if must_exist:
            self._test()

    def _open(self, interface: _RADOSInterface) -> None:
        # TODO: connection caching
        self._api = interface.api
        self._conn = interface.Rados()
        self._conn.connect()
        self._connected = True
        self._ioctx = self._conn.open_ioctx(self._pool)
        self._ioctx.set_namespace(self._ns)
        self._closed = False
        self._offset = 0

    def _test(self) -> None:
        self._ioctx.stat(self._key)

    def read(self, size: typing.Optional[int] = None) -> bytes:
        if self._closed:
            raise ValueError("can not read from closed response")
        return self._read_all() if size is None else self._read(size)

    def _read_all(self) -> bytes:
        ba = bytearray()
        while True:
            chunk = self._read(_CHUNK_SIZE)
            ba += chunk
            if len(chunk) < _CHUNK_SIZE:
                break
        return bytes(ba)

    def _read(self, size: int) -> bytes:
        result = self._ioctx.read(self._key, size, self._offset)
        self._offset += len(result)
        return result

    def close(self) -> None:
        if not self._closed:
            self._ioctx.close()
            self._closed = True
        if self._connected:
            self._conn.shutdown()

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def mode(self) -> str:
        return "rb"

    @property
    def name(self) -> str:
        return self._key

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> None:
        self.close()

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> bytes:
        res = self.read(_CHUNK_SIZE)
        if not res:
            raise StopIteration()
        return res

    def seekable(self) -> bool:
        return False

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def flush(self) -> None:
        pass

    def isatty(self) -> bool:
        return False

    def tell(self) -> int:
        return self._offset

    def seek(self, offset: int, whence: int = 0) -> int:
        raise NotImplementedError()

    def fileno(self) -> int:
        raise NotImplementedError()

    def readline(self, limit: int = -1) -> bytes:
        raise NotImplementedError()

    def readlines(self, hint: int = -1) -> list[bytes]:
        raise NotImplementedError()

    def truncate(self, size: typing.Optional[int] = None) -> int:
        raise NotImplementedError()

    def write(self, s: typing.Any) -> int:
        raise NotImplementedError()

    def writelines(self, ls: typing.Iterable[typing.Any]) -> None:
        raise NotImplementedError()

    def write_full(self, data: bytes) -> None:
        """Write the object such that its contents are exactly `data`."""
        self._ioctx.write_full(self._key, data)

    def _lock_exclusive(self, name: str, cookie: str) -> None:
        self._ioctx.lock_exclusive(
            self._key,
            name,
            cookie,
            desc=self._lock_description,
            duration=self._lock_duration,
        )

    def _acquire_lock_exclusive(
        self, name: str, cookie: str, *, delay: int = 1
    ) -> None:
        while True:
            try:
                self._lock_exclusive(name, cookie)
                return
            except self._api.ObjectBusy:
                _logger.debug(
                    "lock failed: %r, %r, %r: object busy",
                    self._key,
                    name,
                    cookie,
                )
                time.sleep(delay)

    def _unlock(self, name: str, cookie: str) -> None:
        self._ioctx.unlock(self._key, name, cookie)


def _get_mon_config_key(interface: _RADOSInterface, key: str) -> io.BytesIO:
    mcmd = json.dumps(
        {
            "prefix": "config-key get",
            "key": str(key),
        }
    )
    with interface.Rados() as rc:
        ret, out, err = rc.mon_command(mcmd, b"")
        if ret == 0:
            # We need to return a file like object. Since we are handed just
            # bytes from this api, use BytesIO to adapt it to something valid.
            return io.BytesIO(out)
        # ensure ceph didn't send us a negative errno
        ret = ret if ret > 0 else -ret
        msg = f"failed to get mon config key: {key!r}: {err}"
        raise OSError(ret, msg)


class ClusterMetaRADOSHandle:
    "A Cluster Meta Object can load or dump persistent cluster descriptions."

    def __init__(
        self,
        rados_obj: RADOSObjectRef,
        uri: str,
        *,
        read: bool,
        write: bool,
        locked: bool,
    ):
        self._rados_obj = rados_obj
        self._uri = uri
        self._read = read
        self._write = write
        self._locked = locked
        if self._locked:
            self._lock_name = "cluster_meta"
            self._cookie = f"sambacc:{uuid.uuid4()}"

    def load(self) -> typing.Any:
        if not self._read:
            raise ValueError("not readable")
        buf = self._rados_obj.read()
        if not buf:
            return {}
        return json.loads(buf)

    def dump(self, data: typing.Any) -> None:
        if not self._read:
            raise ValueError("not writable")
        buf = json.dumps(data).encode("utf8")
        self._rados_obj.write_full(buf)

    def __enter__(self) -> Self:
        if self._locked:
            self._rados_obj._acquire_lock_exclusive(
                self._lock_name, self._cookie
            )
        return self

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> None:
        if self._locked:
            self._rados_obj._unlock(self._lock_name, self._cookie)
        return


class ClusterMetaRADOSObject:
    def __init__(self, rados_handler: _RADOSHandler, uri: str) -> None:
        self._handler = rados_handler
        self._uri = uri

    def open(
        self, *, read: bool = True, write: bool = False, locked: bool = False
    ) -> ClusterMetaRADOSHandle:
        return ClusterMetaRADOSHandle(
            self._handler.get_object(self._uri),
            self._uri,
            read=read,
            write=write,
            locked=locked,
        )

    @classmethod
    def create_from_uri(cls, uri: str) -> Self:
        """Return a new ClusterMetaRADOSObject given a rados uri string.
        If rados module is unavailable RADOSUnsupported will be raised.
        """
        handler = _RADOSHandler()
        if not handler._interface:
            raise RADOSUnsupported()
        return cls(handler, uri)


def is_rados_uri(uri: str) -> bool:
    """Return true if the string can be used as a rados (pseudo) URI.
    This function does not require the rados libraries to be available.
    NB: It does not validate the structure of the URI.
    """
    return uri.startswith("rados:")


def parse_rados_uri(
    uri: typing.Union[str, urllib.request.Request]
) -> dict[str, str]:
    """Given a rados uri-like value return a dict containing a breakdown of the
    components of the uri.
    """
    req = uri if not isinstance(uri, str) else urllib.request.Request(uri)
    subtype = "mon-config-key"
    if req.selector.startswith(subtype + ":"):
        return {
            "type": req.type,
            "subtype": subtype,
            "path": req.selector.split(":", 1)[1],
        }
    sel = req.selector.lstrip("/")
    if req.host:
        pool = req.host
        ns, key = sel.split("/", 1)
    else:
        pool, ns, key = sel.split("/", 2)
    return {
        "type": req.type,
        "subtype": "object",
        "pool": pool,
        "ns": ns,
        "key": key,
    }


def enable_rados(
    cls: typing.Type[url_opener.URLOpener],
    *,
    client_name: str = "",
    full_name: bool = False,
) -> None:
    """Enable Ceph RADOS support in sambacc.
    As as side-effect it will extend the URLOpener type to support pseudo-URLs
    for rados object storage. It will also enable the
    ClusterMetaRADOSObject.create_from_uri constructor. If rados libraries are
    not found the function does nothing.

    If rados libraries are found than URLOpener can be used like:
    >>> uo = url_opener.URLOpener()
    >>> res = uo.open("rados://my_pool/namepace/obj_key")
    >>> res.read()
    """
    try:
        import rados  # type: ignore[import]
    except ImportError:
        _logger.debug("Failed to import ceph 'rados' module")
        return

    _logger.debug(
        "Enabling ceph rados support with"
        f" client_name={client_name!r}, full_name={full_name}"
    )
    rados_interface = _RADOSInterface()
    rados_interface.api = rados
    rados_interface.client_name = client_name
    rados_interface.full_name = full_name

    _RADOSHandler._interface = rados_interface
    cls._handlers.append(_RADOSHandler)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import pytest

import sambacc.smbconf_api
import sambacc.smbconf_samba

smb_conf_reg_stub = """
[global]
cache directory = {path}
state directory = {path}
private dir = {path}
include = registry
"""

smb_conf_sample = """
[global]
  realm = my.kingdom.fora.horse

[share_a]
  path = /foo/bar/baz
  read only = no
[share_b]
  path = /foo/x/b
  read only = no
[share_c]
  path = /foo/x/c
  read only = no
[share_d]
  path = /foo/x/d
  read only = no
[share_e]
  path = /foo/x/e
  read only = no
"""


def _import_probe():
    try:
        import samba.smbconf  # type: ignore
        import samba.samba3.smbconf  # type: ignore # noqa
    except ImportError:
        pytest.skip("unable to load samba smbconf modules")


def _smb_data(path, smb_conf_text):
    data_path = path / "_samba"
    data_path.mkdir()
    smb_conf_path = path / "smb.conf"
    smb_conf_path.write_text(smb_conf_text.format(path=data_path))
    return smb_conf_path


@pytest.fixture(scope="session")
def smbconf_reg_once(tmp_path_factory):
    _import_probe()
    tmp_path = tmp_path_factory.mktemp("smb_reg")
    smb_conf_path = _smb_data(tmp_path, smb_conf_reg_stub)

    return sambacc.smbconf_samba.SMBConf.from_registry(str(smb_conf_path))


@pytest.fixture(scope="function")
def smbconf_reg(smbconf_reg_once):
    # IMPORTANT: Reminder, samba doesn't release the registry db once opened.
    smbconf_reg_once._smbconf.drop()
    return smbconf_reg_once


@pytest.fixture(scope="function")
def smbconf_file(tmp_path):
    _import_probe()
    smb_conf_path = _smb_data(tmp_path, smb_conf_sample)

    return sambacc.smbconf_samba.SMBConf.from_file(str(smb_conf_path))


def test_smbconf_file_read(smbconf_file):
    assert smbconf_file["global"] == [("realm", "my.kingdom.fora.horse")]
    assert smbconf_file["share_a"] == [
        ("path", "/foo/bar/baz"),
        ("read only", "no"),
    ]
    with pytest.raises(KeyError):
        smbconf_file["not_there"]
    assert list(smbconf_file) == [
        "global",
        "share_a",
        "share_b",
        "share_c",
        "share_d",
        "share_e",
    ]


def test_smbconf_write(smbconf_file):
    assert not smbconf_file.writeable
    with pytest.raises(Exception):
        smbconf_file.import_smbconf(sambacc.smbconf_api.SimpleConfigStore())


def test_smbconf_reg_write_read(smbconf_reg):
    assert smbconf_reg.writeable
    assert list(smbconf_reg) == []
    smbconf_reg["global"] = [("test:one", "1"), ("test:two", "2")]
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "2")]
    smbconf_reg["global"] = [("test:one", "1"), ("test:two", "22")]
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "22")]


def test_smbconf_reg_write_txn_read(smbconf_reg):
    assert smbconf_reg.writeable
    assert list(smbconf_reg) == []
    with smbconf_reg:
        smbconf_reg["global"] = [("test:one", "1"), ("test:two", "2")]
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "2")]
    with smbconf_reg:
        smbconf_reg["global"] = [("test:one", "1"), ("test:two", "22")]
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "22")]

    # transaction with error
    with pytest.raises(ValueError):
        with smbconf_reg:
            smbconf_reg["global"] = [("test:one", "1"), ("test:two", "2222")]
            raise ValueError("foo")
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "22")]

    # no transaction with error
    with pytest.raises(ValueError):
        smbconf_reg["global"] = [("test:one", "1"), ("test:two", "2222")]
        raise ValueError("foo")
    assert smbconf_reg["global"] == [("test:one", "1"), ("test:two", "2222")]


def test_smbconf_reg_import_batched(smbconf_reg, smbconf_file):
    assert list(smbconf_reg) == []
    smbconf_reg.import_smbconf(smbconf_file, batch_size=4)
    assert smbconf_reg["global"] == [("realm", "my.kingdom.fora.horse")]
    assert smbconf_reg["share_a"] == [
        ("path", "/foo/bar/baz"),
        ("read only", "no"),
    ]
    with pytest.raises(KeyError):
        smbconf_reg["not_there"]
    assert list(smbconf_reg) == [
        "global",
        "share_a",
        "share_b",
        "share_c",
        "share_d",
        "share_e",
    ]


def test_smbconf_reg_import_unbatched(smbconf_reg, smbconf_file):
    assert list(smbconf_reg) == []
    smbconf_reg.import_smbconf(smbconf_file, batch_size=None)
    assert smbconf_reg["global"] == [("realm", "my.kingdom.fora.horse")]
    assert smbconf_reg["share_a"] == [
        ("path", "/foo/bar/baz"),
        ("read only", "no"),
    ]
    with pytest.raises(KeyError):
        smbconf_reg["not_there"]
    assert list(smbconf_reg) == [
        "global",
        "share_a",
        "share_b",
        "share_c",
        "share_d",
        "share_e",
    ]
